import copy
from typing import Any
from typing import Dict
from typing import List


def convert_raw_storage_accounts_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    account_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Converts raw storage account api response into present function format

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        account_name: Azure storage account group name.
        resource_id: Azure Policy Definition resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    translated_resounrce = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "account_name": account_name,
        "location": resource["location"],
    }

    if resource.get("sku") and resource.get("sku").get("name"):
        translated_resounrce["sku_name"] = resource["sku"]["name"]

    if resource.get("sku") and resource.get("sku").get("tier"):
        translated_resounrce["sku_tier"] = resource["sku"]["tier"]

    if "kind" in resource:
        translated_resounrce["account_kind"] = resource["kind"]

    if resource.get("extendedLocation") and resource.get("extendedLocation").get(
        "name"
    ):
        translated_resounrce["edge_zone"] = resource["extendedLocation"].get("name")

    if "identity" in resource:
        identity_payload = convert_raw_to_present_identity(
            hub, identity=resource["identity"]
        )
        translated_resounrce["identity"] = identity_payload

    if "tags" in resource:
        translated_resounrce["tags"] = resource["tags"]

    properties = resource.get("properties")
    if properties:
        if "allowCrossTenantReplication" in properties:
            translated_resounrce["cross_tenant_replication_enabled"] = properties[
                "allowCrossTenantReplication"
            ]

        if "accessTier" in properties:
            translated_resounrce["access_tier"] = properties["accessTier"]

        if "supportsHttpsTrafficOnly" in properties:
            translated_resounrce["enable_https_traffic_only"] = properties[
                "supportsHttpsTrafficOnly"
            ]

        if "minimumTlsVersion" in properties:
            translated_resounrce["min_tls_version"] = properties["minimumTlsVersion"]

        if "allowBlobPublicAccess" in properties:
            translated_resounrce["allow_blob_public_access"] = properties[
                "allowBlobPublicAccess"
            ]

        if "allowSharedKeyAccess" in properties:
            translated_resounrce["allow_shared_key_access"] = properties[
                "allowSharedKeyAccess"
            ]

        if "publicNetworkAccess" in properties:
            translated_resounrce["public_network_access"] = properties[
                "publicNetworkAccess"
            ]

        if "defaultToOAuthAuthentication" in properties:
            translated_resounrce["default_to_oauth_authentication"] = properties[
                "defaultToOAuthAuthentication"
            ]

        if "isHnsEnabled" in properties:
            translated_resounrce["is_hns_enabled"] = properties["isHnsEnabled"]

        if "isNfsV3Enabled" in properties:
            translated_resounrce["nfsv3_enabled"] = properties["isNfsV3Enabled"]

        if properties.get("customDomain") is not None:
            customDomainPayload = convert_raw_to_present_custom_domain(
                hub, c_domain=properties.get("customDomain")
            )
            translated_resounrce["custom_domain"] = customDomainPayload

        if properties.get("encryption") is not None:
            customerManagedKeyPayload = convert_raw_to_present_customer_managed_key(
                hub, customer_key=properties.get("encryption")
            )
            if len(customerManagedKeyPayload) != 0:
                translated_resounrce["customer_managed_key"] = customerManagedKeyPayload

        if properties.get("networkAcls") is not None:
            networkRulesPayload = convert_raw_to_present_network_rules(
                hub, rules=properties.get("networkAcls")
            )
            translated_resounrce["network_rules"] = networkRulesPayload

        if "largeFileSharesState" in properties:
            translated_resounrce["large_file_shares_state"] = properties[
                "largeFileSharesState"
            ]

        if properties.get("azureFilesIdentityBasedAuthentication") is not None:
            authPayload = convert_raw_to_present_files_authentication(
                hub, file_auth=properties.get("azureFilesIdentityBasedAuthentication")
            )
            translated_resounrce["azure_files_authentication"] = authPayload

        if properties.get("routingPreference") is not None:
            routing_payload = convert_raw_to_present_routing(
                hub, routing_pref=properties.get("routingPreference")
            )
            translated_resounrce["routing"] = routing_payload

        if properties.get("encryption") is not None:
            encryptionServicePayload = convert_raw_to_present_encryption_service(
                hub, encryption_service=properties.get("encryption")
            )
            translated_resounrce["encryption_service"] = encryptionServicePayload

        if (properties.get("encryption") is not None) and (
            properties.get("encryption").get("requireInfrastructureEncryption")
            is not None
        ):
            translated_resounrce["require_infrastructure_encryption"] = properties[
                "encryption"
            ]["requireInfrastructureEncryption"]

        if (properties.get("immutableStorageWithVersioning") is not None) and (
            properties.get("immutableStorageWithVersioning").get("immutabilityPolicy")
            is not None
        ):
            immutableStoragePayload = convert_raw_to_present_immutability_policy(
                hub,
                immutePolicy=properties.get("immutableStorageWithVersioning").get(
                    "immutabilityPolicy"
                ),
            )
            translated_resounrce["immutability_policy"] = immutableStoragePayload

        if properties.get("sasPolicy") is not None:
            sasPolicyPayload = convert_raw_to_present_sas_policy(
                hub, policy=properties.get("sasPolicy")
            )
            translated_resounrce["sas_policy"] = sasPolicyPayload

        if properties.get("keyPolicy") is not None:
            key_policy_payload = convert_raw_to_present_key_policy(
                hub, policy=properties.get("keyPolicy")
            )
            translated_resounrce["key_policy"] = key_policy_payload

        if "allowedCopyScope" in properties:
            translated_resounrce["allowed_copy_scope"] = properties["allowedCopyScope"]

        if "isSftpEnabled" in properties:
            translated_resounrce["sftp_enabled"] = properties["isSftpEnabled"]

        return translated_resounrce


def convert_present_to_raw_storage_accounts(
    hub,
    location: str,
    sku_name: str,
    sku_tier: str = None,
    account_kind: str = None,
    cross_tenant_replication_enabled: bool = None,
    access_tier: str = None,
    edge_zone: str = None,
    enable_https_traffic_only: bool = None,
    min_tls_version: str = None,
    allow_blob_public_access: bool = None,
    allow_shared_key_access: bool = None,
    public_network_access: str = None,
    default_to_oauth_authentication: bool = None,
    is_hns_enabled: bool = None,
    nfsv3_enabled: bool = None,
    custom_domain: Dict[str, Any] = None,
    customer_managed_key: Dict[str, Any] = None,
    identity: Dict[str, Any] = None,
    network_rules: Dict[str, Any] = None,
    large_file_shares_state: str = None,
    azure_files_authentication: Dict[str, Any] = None,
    routing: Dict[str, Any] = None,
    encryption_service: Dict[str, Any] = None,
    require_infrastructure_encryption: bool = None,
    immutability_policy: Dict[str, Any] = None,
    sas_policy: Dict[str, Any] = None,
    key_policy: Dict[str, Any] = None,
    allowed_copy_scope: str = None,
    sftp_enabled: bool = None,
    tags: Dict[str, str] = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location: Specifies the supported Azure location where the resource exists.
         sku_tier: The SKU name
         sku_name: The SKU tier
         account_kind: Defines the Kind of account.
         cross_tenant_replication_enabled: Allow or disallow cross AAD tenant object replication
         access_tier: Required for storage accounts where kind = BlobStorage. The access tier is used for billing
         edge_zone: Specifies the Edge Zone within the Azure Region where this Storage Account should exist.
         enable_https_traffic_only: Boolean flag which forces HTTPS if enabled.
         min_tls_version: The minimum supported TLS version for the storage account.
         allow_blob_public_access:  Allow or disallow nested items within this Account to opt into being public.
         allow_shared_key_access: Indicates whether the storage account permits requests to be authorized with the account access key via Shared Key.
         public_network_access: Whether the public network access is enabled
         default_to_oauth_authentication: Default to Azure Active Directory authorization in the Azure portal when accessing the Storage Account.
         is_hns_enabled: Account HierarchicalNamespace enabled if sets to true.
         nfsv3_enabled: NFS 3.0 protocol support enabled if set to true.
         custom_domain: User domain assigned to the storage account.
         customer_managed_key: Combination of Key vault key id and user assigned id
         identity: The identity of the resource.
         network_rules: Network rule set
         large_file_shares_state: Allow large file shares if sets to enabled.
         azure_files_authentication: Provides the identity based authentication settings for Azure Files.
         routing: Maintains information about the network routing choice opted by the user for data transfer
         encryption_service: set of encryption services
         require_infrastructure_encryption: A boolean indicating whether or not the service applies a secondary layer of encryption with platform managed keys for data at rest.
         immutability_policy: This argument specifies the default account-level immutability policy which is inherited and applied to objects
         sas_policy: SasPolicy assigned to the storage account.
         key_policy: KeyPolicy assigned to the storage account.
         allowed_copy_scope: Restrict copy to and from Storage Accounts within an AAD tenant or with Private Links to the same VNet.
         sftp_enabled: Enables Secure File Transfer Protocol, if set to true
         tags: Gets or sets a list of key value pairs that describe the resource.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location}

    if tags is not None:
        payload["tags"] = tags

    if (sku_name is not None) or (sku_tier is not None):
        payload["sku"] = {}
    if sku_name is not None:
        payload["sku"]["name"] = sku_name
    if sku_tier is not None:
        payload["sku"]["tier"] = sku_tier

    if account_kind is not None:
        payload["kind"] = account_kind

    if edge_zone is not None:
        payload["extendedLocation"] = {}
        payload["extendedLocation"]["name"] = edge_zone

    if identity is not None:
        payload["identity"] = convert_present_to_raw_identity(hub, identity=identity)

    if (
        (cross_tenant_replication_enabled is not None)
        or (access_tier is not None)
        or (enable_https_traffic_only is not None)
        or (min_tls_version is not None)
        or (allow_blob_public_access is not None)
        or (allow_shared_key_access is not None)
        or (public_network_access is not None)
        or (default_to_oauth_authentication is not None)
        or (is_hns_enabled is not None)
        or (nfsv3_enabled is not None)
        or (custom_domain is not None)
        or (customer_managed_key is not None)
        or (network_rules is not None)
        or (large_file_shares_state is not None)
        or (azure_files_authentication is not None)
        or (routing is not None)
        or (encryption_service is not None)
        or (require_infrastructure_encryption is not None)
        or (immutability_policy is not None)
        or (sas_policy is not None)
        or (key_policy is not None)
        or (allowed_copy_scope is not None)
        or (sftp_enabled is not None)
    ):
        payload["properties"] = {}
    if cross_tenant_replication_enabled is not None:
        payload["properties"][
            "allowCrossTenantReplication"
        ] = cross_tenant_replication_enabled

    if access_tier is not None:
        payload["properties"]["accessTier"] = access_tier

    if enable_https_traffic_only is not None:
        payload["properties"]["supportsHttpsTrafficOnly"] = enable_https_traffic_only

    if min_tls_version is not None:
        payload["properties"]["minimumTlsVersion"] = min_tls_version

    if allow_blob_public_access is not None:
        payload["properties"]["allowBlobPublicAccess"] = allow_blob_public_access

    if allow_shared_key_access is not None:
        payload["properties"]["allowSharedKeyAccess"] = allow_shared_key_access

    if public_network_access is not None:
        payload["properties"]["publicNetworkAccess"] = public_network_access

    if default_to_oauth_authentication is not None:
        payload["properties"][
            "defaultToOAuthAuthentication"
        ] = default_to_oauth_authentication

    if is_hns_enabled is not None:
        payload["properties"]["isHnsEnabled"] = is_hns_enabled

    if nfsv3_enabled is not None:
        payload["properties"]["isNfsV3Enabled"] = nfsv3_enabled

    if custom_domain is not None:
        payload["properties"]["customDomain"] = convert_present_to_raw_custom_domain(
            hub, c_domain=custom_domain
        )

    if (
        (customer_managed_key is not None)
        or (encryption_service is not None)
        or (require_infrastructure_encryption is not None)
    ):
        payload["properties"]["encryption"] = {}

    if customer_managed_key is not None:
        payload["properties"]["encryption"]["keyvaultproperties"] = {}

        if customer_managed_key.get("key_vault_key_id") is not None:
            payload["properties"]["encryption"]["keyvaultproperties"][
                "currentVersionedKeyIdentifier"
            ] = customer_managed_key["key_vault_key_id"]

        if customer_managed_key.get("key_name") is not None:
            payload["properties"]["encryption"]["keyvaultproperties"][
                "keyname"
            ] = customer_managed_key.get("key_name")

        if customer_managed_key.get("key_version") is not None:
            payload["properties"]["encryption"]["keyvaultproperties"][
                "keyversion"
            ] = customer_managed_key.get("key_version")

        if customer_managed_key.get("key_vault_uri") is not None:
            payload["properties"]["encryption"]["keyvaultproperties"][
                "keyvaulturi"
            ] = customer_managed_key.get("key_vault_uri")

        payload["properties"]["encryption"]["identity"] = {}
        payload["properties"]["encryption"]["identity"][
            "userAssignedIdentity"
        ] = customer_managed_key["user_assigned_identity_id"]

        if customer_managed_key.get("federated_identity_client_id") is not None:
            payload["properties"]["encryption"]["identity"][
                "federatedIdentityClientId"
            ] = customer_managed_key.get("federated_identity_client_id")

    if network_rules is not None:
        payload["properties"]["networkAcls"] = convert_present_to_raw_network_rules(
            hub, rules=network_rules
        )

    if large_file_shares_state is not None:
        payload["properties"]["largeFileSharesState"] = large_file_shares_state

    if azure_files_authentication is not None:
        payload["properties"][
            "azureFilesIdentityBasedAuthentication"
        ] = convert_present_to_raw_files_authentication(
            hub, file_auth=azure_files_authentication
        )

    if routing is not None:
        payload["properties"]["routingPreference"] = convert_present_to_raw_routing(
            hub, routing_pref=routing
        )

    if encryption_service is not None:
        payload["properties"]["encryption"]["keySource"] = encryption_service[
            "encryption_key_source"
        ]
        if (
            (encryption_service.get("queue_encryption_key_type") is not None)
            or (encryption_service.get("table_encryption_key_type") is not None)
            or (encryption_service.get("blob_encryption_key_type") is not None)
            or (encryption_service.get("file_encryption_key_type") is not None)
        ):
            payload["properties"]["encryption"]["services"] = {}

        if encryption_service.get("queue_encryption_key_type") is not None:
            payload["properties"]["encryption"]["services"]["queue"] = {}
            payload["properties"]["encryption"]["services"]["queue"][
                "keyType"
            ] = encryption_service.get("queue_encryption_key_type")

        if encryption_service.get("table_encryption_key_type") is not None:
            payload["properties"]["encryption"]["services"]["table"] = {}
            payload["properties"]["encryption"]["services"]["table"][
                "keyType"
            ] = encryption_service.get("table_encryption_key_type")

        if encryption_service.get("blob_encryption_key_type") is not None:
            payload["properties"]["encryption"]["services"]["blob"] = {}
            payload["properties"]["encryption"]["services"]["blob"][
                "keyType"
            ] = encryption_service.get("blob_encryption_key_type")

        if encryption_service.get("file_encryption_key_type") is not None:
            payload["properties"]["encryption"]["services"]["file"] = {}
            payload["properties"]["encryption"]["services"]["file"][
                "keyType"
            ] = encryption_service.get("file_encryption_key_type")

    if require_infrastructure_encryption is not None:
        payload["properties"]["encryption"][
            "requireInfrastructureEncryption"
        ] = require_infrastructure_encryption

    if immutability_policy is not None:
        payload["properties"]["immutableStorageWithVersioning"] = {}
        payload["properties"]["immutableStorageWithVersioning"][
            "immutabilityPolicy"
        ] = convert_present_to_raw_immutability_policy(
            hub, immute_policy=immutability_policy
        )

    if sas_policy is not None:
        payload["properties"]["sasPolicy"] = convert_present_to_raw_sas_policy(
            hub, policy=sas_policy
        )

    if key_policy is not None:
        payload["properties"]["keyPolicy"] = convert_present_to_raw_key_policy(
            hub, policy=key_policy
        )

    if allowed_copy_scope is not None:
        payload["properties"]["allowedCopyScope"] = allowed_copy_scope

    if sftp_enabled is not None:
        payload["properties"]["isSftpEnabled"] = sftp_enabled

    return payload


def update_storage_accounts_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)

    if (
        (new_values.get("sku_tier") is not None)
        and (existing_payload.get("sku") is not None)
        and (new_values["sku_tier"] != existing_payload.get("sku").get("tier"))
    ):
        new_payload["sku"]["tier"] = new_values.get("sku_tier")
        need_update = True

    if (
        (new_values.get("sku_name") is not None)
        and (existing_payload.get("sku") is not None)
        and (new_values["sku_name"] != existing_payload.get("sku").get("name"))
    ):
        new_payload["sku"]["name"] = new_values.get("sku_name")
        need_update = True

    if (new_values.get("account_kind") is not None) and (
        new_values["account_kind"] != existing_payload.get("kind")
    ):
        new_payload["kind"] = new_values.get("account_kind")
        need_update = True

    if (
        (new_values.get("cross_tenant_replication_enabled") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["cross_tenant_replication_enabled"]
            != existing_payload.get("properties").get("allowCrossTenantReplication")
        )
    ):
        new_payload["properties"]["allowCrossTenantReplication"] = new_values.get(
            "cross_tenant_replication_enabled"
        )
        need_update = True

    if (
        (new_values.get("access_tier") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["access_tier"]
            != existing_payload.get("properties").get("accessTier")
        )
    ):
        new_payload["properties"]["accessTier"] = new_values.get("access_tier")
        need_update = True

    if (
        (new_values.get("edge_zone") is not None)
        and (existing_payload.get("extendedLocation") is not None)
        and (
            new_values["edge_zone"]
            != existing_payload.get("extendedLocation").get("name")
        )
    ):
        new_payload["extendedLocation"]["type"] = new_values.get("edge_zone")
        need_update = True

    if (
        (new_values.get("enable_https_traffic_only") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["enable_https_traffic_only"]
            != existing_payload.get("properties").get("supportsHttpsTrafficOnly")
        )
    ):
        new_payload["properties"]["supportsHttpsTrafficOnly"] = new_values.get(
            "enable_https_traffic_only"
        )
        need_update = True

    if (
        (new_values.get("min_tls_version") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["min_tls_version"]
            != existing_payload.get("properties").get("minimumTlsVersion")
        )
    ):
        new_payload["properties"]["minimumTlsVersion"] = new_values.get(
            "min_tls_version"
        )
        need_update = True

    if (
        (new_values.get("allow_blob_public_access") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["allow_blob_public_access"]
            != existing_payload.get("properties").get("allowBlobPublicAccess")
        )
    ):
        new_payload["properties"]["allowBlobPublicAccess"] = new_values.get(
            "allow_blob_public_access"
        )
        need_update = True

    if (
        (new_values.get("allow_shared_key_access") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["allow_shared_key_access"]
            != existing_payload.get("properties").get("allowSharedKeyAccess")
        )
    ):
        new_payload["properties"]["allowSharedKeyAccess"] = new_values.get(
            "allow_shared_key_access"
        )
        need_update = True

    if (
        (new_values.get("public_network_access") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["public_network_access"]
            != existing_payload.get("properties").get("publicNetworkAccess")
        )
    ):
        new_payload["properties"]["publicNetworkAccess"] = new_values.get(
            "public_network_access"
        )
        need_update = True

    if (
        (new_values.get("default_to_oauth_authentication") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["default_to_oauth_authentication"]
            != existing_payload.get("properties").get("defaultToOAuthAuthentication")
        )
    ):
        new_payload["properties"]["defaultToOAuthAuthentication"] = new_values.get(
            "default_to_oauth_authentication"
        )
        need_update = True

    if (
        (new_values.get("is_hns_enabled") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["is_hns_enabled"]
            != existing_payload.get("properties").get("isHnsEnabled")
        )
    ):
        new_payload["properties"]["isHnsEnabled"] = new_values.get("is_hns_enabled")
        need_update = True

    if (
        (new_values.get("nfsv3_enabled") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["nfsv3_enabled"]
            != existing_payload.get("properties").get("isNfsV3Enabled")
        )
    ):
        new_payload["properties"]["isNfsV3Enabled"] = new_values.get("nfsv3_enabled")
        need_update = True

    if (new_values.get("custom_domain") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_custom_domain = existing_payload.get("properties").get(
            "customDomain", {}
        )
        existing_custom_domain_required_payload = convert_raw_to_present_custom_domain(
            hub, c_domain=existing_custom_domain
        )
        if compare_update_dict_payload(
            existing_custom_domain_required_payload, new_values.get("custom_domain")
        ):
            new_payload["properties"][
                "customDomain"
            ] = convert_present_to_raw_custom_domain(
                hub, c_domain=new_values.get("custom_domain")
            )
            need_update = True
    if (new_values.get("customer_managed_key") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_customer_managed_key = existing_payload.get("properties").get(
            "encryption", {}
        )
        existing_customer_managed_key_required_payload = (
            convert_raw_to_present_customer_managed_key(
                hub, customer_key=existing_customer_managed_key
            )
        )
        if compare_update_dict_payload(
            existing_customer_managed_key_required_payload,
            new_values.get("customer_managed_key"),
        ):
            if (
                new_values.get("customer_managed_key", {}).get(
                    "user_assigned_identity_id"
                )
                is not None
            ):
                new_payload["properties"]["encryption"]["identity"][
                    "userAssignedIdentity"
                ] = new_values.get("customer_managed_key").get(
                    "user_assigned_identity_id"
                )

            if (
                new_values.get("customer_managed_key", {}).get("key_vault_key_id")
                is not None
            ):
                new_payload["properties"]["encryption"]["keyvaultproperties"] = {
                    "currentVersionedKeyIdentifier": new_values.get(
                        "customer_managed_key"
                    ).get("key_vault_key_id"),
                }

            if new_values.get("customer_managed_key", {}).get("key_name") is not None:
                new_payload["properties"]["encryption"]["keyvaultproperties"] = {
                    "keyname": new_values.get("customer_managed_key").get("key_name"),
                }

            if (
                new_values.get("customer_managed_key", {}).get("key_version")
                is not None
            ):
                new_payload["properties"]["encryption"]["keyvaultproperties"] = {
                    "keyversion": new_values.get("customer_managed_key").get(
                        "key_version"
                    ),
                }

            if (
                new_values.get("customer_managed_key", {}).get("key_vault_uri")
                is not None
            ):
                new_payload["properties"]["encryption"]["keyvaultproperties"] = {
                    "keyvaulturi": new_values.get("customer_managed_key").get(
                        "key_vault_uri"
                    ),
                }
            need_update = True

    if new_values.get("identity") is not None:
        existing_identity = existing_payload.get("identity", {})
        existing_identity_required_payload = convert_raw_to_present_identity(
            hub, identity=existing_identity
        )
        if compare_update_dict_payload(
            existing_identity_required_payload, new_values.get("identity")
        ):
            new_payload["identity"] = convert_present_to_raw_identity(
                hub, identity=new_values.get("identity")
            )
            need_update = True

    if (new_values.get("network_rules") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_network_rules = existing_payload.get("properties").get(
            "networkAcls", {}
        )
        existing_network_rules_required_payload = convert_raw_to_present_network_rules(
            hub, rules=existing_network_rules
        )
        if compare_update_dict_payload(
            existing_network_rules_required_payload, new_values.get("network_rules")
        ):
            new_payload["properties"][
                "networkAcls"
            ] = convert_present_to_raw_network_rules(
                hub, rules=new_values.get("network_rules")
            )
            need_update = True

    if (
        (new_values.get("large_file_shares_state") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["large_file_shares_state"]
            != existing_payload.get("properties").get("largeFileSharesState")
        )
    ):
        new_payload["properties"]["largeFileSharesState"] = new_values.get(
            "large_file_shares_state"
        )
        need_update = True

    if (new_values.get("azure_files_authentication") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_files_authentication = existing_payload.get("properties").get(
            "azureFilesIdentityBasedAuthentication", {}
        )
        existing_files_authentication_required_payload = (
            convert_raw_to_present_files_authentication(
                hub, file_auth=existing_files_authentication
            )
        )
        if compare_update_dict_payload(
            existing_files_authentication_required_payload,
            new_values.get("azure_files_authentication"),
        ):
            new_payload["properties"][
                "azureFilesIdentityBasedAuthentication"
            ] = convert_present_to_raw_files_authentication(
                hub, file_auth=new_values.get("azure_files_authentication")
            )
            need_update = True

    if (new_values.get("routing") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_routing = existing_payload.get("properties").get(
            "routingPreference", {}
        )
        existing_routing_required_payload = convert_raw_to_present_routing(
            hub, routing_pref=existing_routing
        )
        if compare_update_dict_payload(
            existing_routing_required_payload, new_values.get("routing")
        ):
            new_payload["properties"][
                "routingPreference"
            ] = convert_present_to_raw_routing(
                hub, routing_pref=new_values.get("routing")
            )
            need_update = True

    if (new_values.get("encryption_service") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_encryption_service = existing_payload.get("properties").get(
            "encryption", {}
        )
        existing_encryption_service_required_payload = (
            convert_raw_to_present_encryption_service(
                hub, encryption_service=existing_encryption_service
            )
        )
        if compare_update_dict_payload(
            existing_encryption_service_required_payload,
            new_values.get("encryption_service"),
        ):
            new_payload["properties"]["encryption"]["keySource"] = new_values.get(
                "encryption_service"
            ).get("encryption_key_source")
            if (
                existing_payload.get("properties").get("encryption").get("services")
                is not None
            ):
                new_services = {}
                if (
                    new_values.get("encryption_service").get(
                        "queue_encryption_key_type"
                    )
                    is not None
                ):
                    new_services["queue"] = {
                        "keyType": new_values.get("encryption_service").get(
                            "queue_encryption_key_type"
                        )
                    }
                if (
                    new_values.get("encryption_service").get(
                        "table_encryption_key_type"
                    )
                    is not None
                ):
                    new_services["table"] = {
                        "keyType": new_values.get("encryption_service").get(
                            "table_encryption_key_type"
                        ),
                    }
                if (
                    new_values.get("encryption_service").get("blob_encryption_key_type")
                    is not None
                ):
                    new_services["blob"] = {
                        "keyType": new_values.get("encryption_service").get(
                            "blob_encryption_key_type"
                        ),
                    }
                if (
                    new_values.get("encryption_service").get("file_encryption_key_type")
                    is not None
                ):
                    new_services["file"] = {
                        "keyType": new_values.get("encryption_service").get(
                            "file_encryption_key_type"
                        ),
                    }
                if new_services:
                    new_payload["properties"]["encryption"]["services"] = new_services
            need_update = True

    if (
        (new_values.get("require_infrastructure_encryption") is not None)
        and (existing_payload.get("properties") is not None)
        and (existing_payload.get("properties").get("encryption") is not None)
        and (
            new_values["require_infrastructure_encryption"]
            != existing_payload.get("properties")
            .get("encryption")
            .get("requireInfrastructureEncryption")
        )
    ):
        new_payload["properties"]["encryption"][
            "requireInfrastructureEncryption"
        ] = new_values.get("require_infrastructure_encryption")
        need_update = True

    if (
        (new_values.get("immutability_policy") is not None)
        and (existing_payload.get("properties") is not None)
        and (existing_payload.get("immutableStorageWithVersioning") is not None)
    ):
        existing_immutability_policy = (
            existing_payload.get("properties")
            .get("immutableStorageWithVersioning")
            .get("immutabilityPolicy", {})
        )
        existing_immutability_policy_required_payload = (
            convert_raw_to_present_immutability_policy(
                hub, immute_policy=existing_immutability_policy
            )
        )
        if compare_update_dict_payload(
            existing_immutability_policy_required_payload,
            new_values.get("immutability_policy"),
        ):
            new_payload["properties"]["immutableStorageWithVersioning"][
                "immutabilityPolicy"
            ] = convert_present_to_raw_immutability_policy(
                hub, immute_policy=new_values.get("immutability_policy")
            )
            need_update = True

    if (new_values.get("sas_policy") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_sas_policy = existing_payload.get("properties").get("sasPolicy", {})
        existing_sas_policy_required_payload = convert_raw_to_present_sas_policy(
            hub, policy=existing_sas_policy
        )
        if compare_update_dict_payload(
            existing_sas_policy_required_payload, new_values.get("sas_policy")
        ):
            new_payload["properties"]["sasPolicy"] = convert_present_to_raw_sas_policy(
                hub, policy=new_values.get("sas_policy")
            )
            need_update = True

    if (new_values.get("key_policy") is not None) and (
        existing_payload.get("properties") is not None
    ):
        existing_key_policy = existing_payload.get("properties").get("keyPolicy", {})
        existing_key_policy_required_payload = convert_raw_to_present_key_policy(
            hub, policy=existing_key_policy
        )
        if compare_update_dict_payload(
            existing_key_policy_required_payload, new_values.get("key_policy")
        ):
            new_payload["properties"]["keyPolicy"] = convert_present_to_raw_key_policy(
                hub, policy=new_values.get("key_policy")
            )
            need_update = True

    if (
        (new_values.get("allowed_copy_scope") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["allowed_copy_scope"]
            != existing_payload.get("properties").get("allowedCopyScope")
        )
    ):
        new_payload["properties"]["allowedCopyScope"] = new_values.get(
            "allowed_copy_scope"
        )
        need_update = True

    if (
        (new_values.get("sftp_enabled") is not None)
        and (existing_payload.get("properties") is not None)
        and (
            new_values["sftp_enabled"]
            != existing_payload.get("properties").get("isSftpEnabled")
        )
    ):
        new_payload["properties"]["isSftpEnabled"] = new_values.get("sftp_enabled")
        need_update = True

    if (new_values.get("tags") is not None) and (
        new_values["tags"] != existing_payload.get("tags")
    ):
        new_payload["tags"] = new_values.get("tags")
        need_update = True

    if need_update:
        result["ret"] = new_payload
    return result


def compare_update_dict_payload(
    existing_payload: Dict[str, Any], update_payload: Dict[str, Any]
):
    """
    Compares the payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_payload(Dict): Existing payload
        update_payload(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    for parameter in update_payload:
        if parameter in existing_payload:
            if update_payload.get(parameter) != existing_payload.get(parameter):
                return True
        else:
            return True
    return False


def convert_raw_to_present_custom_domain(hub, c_domain: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        c_domain(Dict, optional): Custom Disk payload in a storage account resource.

    Returns:
         Custom Domain payload that contains the parameters that match respective present function's input format.
    """
    custom_domain_payload = {}

    if c_domain.get("name") is not None:
        custom_domain_payload["name"] = c_domain.get("name")

    if c_domain.get("useSubDomainName") is not None:
        custom_domain_payload["use_subdomain"] = c_domain.get("useSubDomainName")

    return custom_domain_payload


def convert_raw_to_present_customer_managed_key(hub, customer_key: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        customer_key(Dict, optional): customer key payload in a storage account resource.

    Returns:
         customer key payload that contains the parameters that match respective present function's input format.
    """
    customer_key_present = {}

    if customer_key.get("keyvaultproperties") is not None:
        customer_key_present["key_vault_key_id"] = customer_key.get(
            "keyvaultproperties"
        ).get("currentVersionedKeyIdentifier")
        customer_key_present["key_name"] = customer_key.get("keyvaultproperties").get(
            "keyname"
        )
        customer_key_present["key_version"] = customer_key.get(
            "keyvaultproperties"
        ).get("keyversion")
        customer_key_present["key_vault_uri"] = customer_key.get(
            "keyvaultproperties"
        ).get("keyvaulturi")

    if customer_key.get("identity") is not None:
        customer_key_present["user_assigned_identity_id"] = customer_key.get(
            "identity"
        ).get("userAssignedIdentity")
        customer_key_present["federated_identity_client_id"] = customer_key.get(
            "identity"
        ).get("federatedIdentityClientId")

    customer_key_present = {
        k: v for k, v in customer_key_present.items() if v is not None
    }

    return customer_key_present


def convert_raw_to_present_identity(hub, identity: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        identity(Dict, optional): Identity payload in a storage account resource.

    Returns:
         Identity payload that contains the parameters that match respective present function's input format.
    """
    identity_payload = {}

    if identity.get("type") is not None:
        identity_payload["type"] = identity.get("type")

    if identity.get("userAssignedIdentities") is not None:
        identity_payload["user_assigned_identities"] = identity.get(
            "userAssignedIdentities"
        ).copy()

    return identity_payload


def convert_raw_to_present_sas_policy(hub, policy: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        policy(Dict, optional): sas policy payload in a storage account resource.

    Returns:
         sas policy payload that contains the parameters that match respective present function's input format.
    """
    sas_policy_payload = {}

    if policy.get("sasExpirationPeriod") is not None:
        sas_policy_payload["expiration_period"] = policy["sasExpirationPeriod"]

    if policy.get("expirationAction") is not None:
        sas_policy_payload["expiration_action"] = policy["expirationAction"]

    return sas_policy_payload


def convert_raw_to_present_key_policy(hub, policy: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        policy(Dict, optional): key policy payload in a storage account resource.

    Returns:
         key policy payload that contains the parameters that match respective present function's input format.
    """
    key_policy_payload = {}

    if policy.get("keyExpirationPeriodInDays") is not None:
        key_policy_payload["key_expiration_period_in_days"] = policy.get(
            "keyExpirationPeriodInDays"
        )

    return key_policy_payload


def convert_raw_to_present_immutability_policy(hub, immutePolicy: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

     Args:
         immutePolicy(Dict, optional): immutability policy payload in a storage account resource.

    Returns:
        immutability policy payload that contains the parameters that match respective present function's input format.
    """
    immutability_policy_payload = {}

    if immutePolicy.get("allowProtectedAppendWrites") is not None:
        immutability_policy_payload["allow_protected_append_writes"] = immutePolicy.get(
            "allowProtectedAppendWrites"
        )

    if immutePolicy.get("state") is not None:
        immutability_policy_payload["state"] = immutePolicy.get("state")

    if immutePolicy.get("immutabilityPeriodSinceCreationInDays") is not None:
        immutability_policy_payload["period_since_creation_in_days"] = immutePolicy.get(
            "immutabilityPeriodSinceCreationInDays"
        )

    return immutability_policy_payload


def convert_raw_to_present_routing(hub, routing_pref: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

     Args:
         routing_pref(Dict, optional): routing payload in a storage account resource.

    Returns:
        routing payload that contains the parameters that match respective present function's input format.
    """
    routing_payload = {}

    if routing_pref.get("publishInternetEndpoints") is not None:
        routing_payload["publish_internet_endpoints"] = routing_pref[
            "publishInternetEndpoints"
        ]

    if routing_pref.get("publishMicrosoftEndpoints") is not None:
        routing_payload["publish_microsoft_endpoints"] = routing_pref[
            "publishMicrosoftEndpoints"
        ]

    if routing_pref.get("routingChoice") is not None:
        routing_payload["routing_choice"] = routing_pref["routingChoice"]
    return routing_payload


def convert_raw_to_present_encryption_service(hub, encryption_service: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        encryption_service(Dict, optional): encryption service payload in a storage account resource.

    Returns:
         encryption service payload that contains the parameters that match respective present function's input format.
    """
    encryption_service_payload = {}
    if encryption_service.get("keySource") is not None:
        encryption_service_payload["encryption_key_source"] = encryption_service.get(
            "keySource"
        )

    if (
        (encryption_service.get("services") is not None)
        and (encryption_service.get("services").get("queue") is not None)
        and (encryption_service.get("services").get("queue").get("keyType") is not None)
    ):
        encryption_service_payload["queue_encryption_key_type"] = (
            encryption_service.get("services").get("queue").get("keyType")
        )

    if (
        (encryption_service.get("services") is not None)
        and (encryption_service.get("services").get("table") is not None)
        and (encryption_service.get("services").get("table").get("keyType") is not None)
    ):
        encryption_service_payload["table_encryption_key_type"] = (
            encryption_service.get("services").get("table").get("keyType")
        )

    if (
        (encryption_service.get("services") is not None)
        and (encryption_service.get("services").get("blob") is not None)
        and (encryption_service.get("services").get("blob").get("keyType") is not None)
    ):
        encryption_service_payload["blob_encryption_key_type"] = (
            encryption_service.get("services").get("blob").get("keyType")
        )

    if (
        (encryption_service.get("services") is not None)
        and (encryption_service.get("services").get("file") is not None)
        and (encryption_service.get("services").get("file").get("keyType") is not None)
    ):
        encryption_service_payload["file_encryption_key_type"] = (
            encryption_service.get("services").get("file").get("keyType")
        )

    return encryption_service_payload


def convert_raw_to_present_files_authentication(hub, file_auth: Dict[str, Any]):
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        file_auth(Dict, optional): file authentication payload in a storage account resource.

    Returns:
        file authentication payload that contains the parameters that match respective present function's input format.
    """
    files_authentication_payload = {}

    if file_auth.get("directoryServiceOptions") is not None:
        files_authentication_payload["directory_service_options"] = file_auth.get(
            "directoryServiceOptions"
        )

    if file_auth.get("defaultSharePermission") is not None:
        files_authentication_payload["default_share_permission"] = file_auth.get(
            "defaultSharePermission"
        )

    if file_auth.get("activeDirectoryProperties") is not None:
        active_directory_payload = covert_raw_to_present_active_directory(
            hub, directory_properties=file_auth.get("activeDirectoryProperties")
        )
        files_authentication_payload[
            "active_directory_properties"
        ] = active_directory_payload
    return files_authentication_payload


def covert_raw_to_present_active_directory(hub, directory_properties: Dict[str, Any]):
    current_directory_payload = {}

    if directory_properties.get("domainName") is not None:
        current_directory_payload["domain_name"] = directory_properties["domainName"]

    if directory_properties.get("domainGuid") is not None:
        current_directory_payload["domain_guid"] = directory_properties["domainGuid"]

    if directory_properties.get("azureStorageSid") is not None:
        current_directory_payload["azure_storage_sid"] = directory_properties[
            "azureStorageSid"
        ]

    if directory_properties.get("domainSid") is not None:
        current_directory_payload["domain_sid"] = directory_properties["domainSid"]

    if directory_properties.get("forestName") is not None:
        current_directory_payload["forest_name"] = directory_properties["forestName"]

    if directory_properties.get("netBiosDomainName") is not None:
        current_directory_payload["netbios_domain_name"] = directory_properties[
            "netBiosDomainName"
        ]

    return current_directory_payload


def convert_raw_to_present_network_rules(hub, rules: Dict[str, Any]):
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        rules(Dict, optional): network rules payload in a storage account resource.

    Returns:
       network rules payload that contains the parameters that match respective present function's input format.
    """
    network_rule_payload = {}

    if rules.get("defaultAction") is not None:
        network_rule_payload["default_action"] = rules.get("defaultAction")

    if rules.get("bypass") is not None:
        network_rule_payload["bypass"] = rules["bypass"]

    if rules.get("ipRules") is not None:
        ip = []
        for ip_rule in rules["ipRules"]:
            ip.append(ip_rule.get("value"))
        network_rule_payload["ip_rule_values"] = ip

    if rules.get("virtualNetworkRules") is not None:
        subnetIds = []
        for vnRules in rules["virtualNetworkRules"]:
            subnetIds.append(vnRules.get("id"))
        network_rule_payload["virtual_network_subnet_ids"] = subnetIds

    if rules.get("resourceAccessRules") is not None:
        resource_access_rules_payload = convert_raw_to_present_resource_access_rules(
            hub, resource_access=rules.get("resourceAccessRules")
        )
        network_rule_payload["resource_access_rules"] = resource_access_rules_payload

    return network_rule_payload


def convert_raw_to_present_resource_access_rules(
    hub, resource_access: List[Dict[str, Any]]
):
    present_resource_access_rule = []

    for access_rule in resource_access:
        payload = {"endpoint_resource_id": access_rule["resourceId"]}

        if access_rule.get("tenantId") is not None:
            payload["endpoint_tenant_id"] = access_rule["tenantId"]
        present_resource_access_rule.append(payload)

    return present_resource_access_rule


def convert_present_to_raw_custom_domain(hub, c_domain: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        c_domain(Dict(str, Any)) :

    Returns:
        Custom Domain Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    custom_domain_payload = {
        "name": c_domain.get("name"),
    }

    if c_domain.get("use_subdomain") is not None:
        custom_domain_payload["useSubDomainName"] = c_domain["use_subdomain"]

    return custom_domain_payload


def convert_present_to_raw_identity(hub, identity: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        identity(Dict(str, Any)) :

    Returns:
        identity Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    identity_payload = {
        "type": identity.get("type"),
    }

    if identity.get("user_assigned_identities") is not None:
        identity_payload["userAssignedIdentities"] = identity.get(
            "user_assigned_identities"
        ).copy()

        return identity_payload


def convert_present_to_raw_sas_policy(hub, policy: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        policy(Dict(str, Any)) :

    Returns:
        Sas Policy Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    sas_policy_payload = {
        "sasExpirationPeriod": policy.get("expiration_period"),
    }

    if policy.get("expiration_action") is not None:
        sas_policy_payload["expirationAction"] = policy["expiration_action"]

    return sas_policy_payload


def convert_present_to_raw_key_policy(hub, policy: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        policy(Dict(str, Any)) :

    Returns:
        Key Policy Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    key_policy_payload = {
        "keyExpirationPeriodInDays": policy.get("key_expiration_period_in_days"),
    }

    return key_policy_payload


def convert_present_to_raw_immutability_policy(hub, immute_policy: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        immute_policy(Dict(str, Any)) :

    Returns:
        Immutability Policy Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    immutability_policy_payload = {
        "allowProtectedAppendWrites": immute_policy["allow_protected_append_writes"],
        "state": immute_policy["state"],
        "immutabilityPeriodSinceCreationInDays": immute_policy[
            "period_since_creation_in_days"
        ],
    }
    return immutability_policy_payload


def convert_present_to_raw_routing(hub, routing_pref: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        routing_pref(Dict(str, Any)) :

    Returns:
        Routing Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    routing_payload = {}

    if routing_pref.get("publish_internet_endpoints") is not None:
        routing_payload["publishInternetEndpoints"] = routing_pref[
            "publish_internet_endpoints"
        ]

    if routing_pref.get("publish_microsoft_endpoints") is not None:
        routing_payload["publishMicrosoftEndpoints"] = routing_pref[
            "publish_microsoft_endpoints"
        ]

    if routing_pref.get("routing_choice") is not None:
        routing_payload["routingChoice"] = routing_pref["routing_choice"]
    return routing_payload


def convert_present_to_raw_files_authentication(hub, file_auth: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
       file_auth(Dict(str, Any)) :

    Returns:
        file authentication Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    files_authentication_payload = {
        "directoryServiceOptions": file_auth["directory_service_options"]
    }

    if file_auth.get("default_share_permission"):
        files_authentication_payload["defaultSharePermission"] = file_auth.get(
            "default_share_permission"
        )

    if file_auth.get("active_directory_properties"):
        files_authentication_payload[
            "activeDirectoryProperties"
        ] = convert_present_to_raw_active_directory(
            hub, directory_properties=file_auth.get("active_directory_properties")
        )
    return files_authentication_payload


def convert_present_to_raw_active_directory(hub, directory_properties: Dict[str, Any]):
    current_directory_payload = {
        "domainName": directory_properties["domain_name"],
        "domainGuid": directory_properties["domain_guid"],
    }

    if directory_properties.get("azure_storage_sid") is not None:
        current_directory_payload["azureStorageSid"] = directory_properties[
            "azure_storage_sid"
        ]

    if directory_properties.get("domain_sid") is not None:
        current_directory_payload["domainSid"] = directory_properties["domain_sid"]

    if directory_properties.get("forest_name") is not None:
        current_directory_payload["forestName"] = directory_properties["forest_name"]

    if directory_properties.get("netbios_domain_name") is not None:
        current_directory_payload["netBiosDomainName"] = directory_properties[
            "netbios_domain_name"
        ]

    return current_directory_payload


def convert_present_to_raw_network_rules(hub, rules: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
       rules(Dict(str, Any)) :

    Returns:
       Network rules Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    network_rule_payload = {
        "defaultAction": rules["default_action"],
    }

    if rules.get("bypass") is not None:
        network_rule_payload["bypass"] = rules["bypass"]

    if rules.get("ip_rule_values") is not None:
        ipRuleList = []
        for value in rules["ip_rule_values"]:
            ipRuleList.append({"action": "Allow", "value": value})
        network_rule_payload["ipRules"] = ipRuleList

    if rules.get("virtual_network_subnet_ids") is not None:
        virtualNetworkRuleList = []
        for id in rules["virtual_network_subnet_ids"]:
            virtualNetworkRuleList.append(
                {"action": "Allow", "id": id, "state": "Succeeded"}
            )
        network_rule_payload["virtualNetworkRules"] = virtualNetworkRuleList

    if rules.get("resource_access_rules") is not None:
        network_rule_payload[
            "resourceAccessRules"
        ] = convert_present_to_raw_resource_access_rules(
            hub, resource_access=rules.get("resource_access_rules")
        )

    return network_rule_payload


def convert_present_to_raw_resource_access_rules(
    hub, resource_access: List[Dict[str, Any]]
):
    raw_access_rules = []

    for access_rule in resource_access:
        payload = {
            "resourceId": access_rule["endpoint_resource_id"],
        }
        if access_rule.get("endpoint_tenant_id") is not None:
            payload["tenantId"] = access_rule["endpoint_tenant_id"]
        raw_access_rules.append(payload)

    return raw_access_rules



from collections import OrderedDict
from typing import Dict


def get_parameter_value(hub, uri: str, parameters: OrderedDict) -> list:
    """
    Generate a list of pairs of uri parameter and its value. And convert uri parameters to expected key used in sls files
    For example /virtualNetworks/virtual-network-name/subnets/subnet-name} will output
    [{"virtualNetworks": "virtual-network-name"}, {"subnets": "subnet-name"}]
    :param hub: The redistributed pop central hub. This is required in Idem, so while not used, must appear.
    :param uri: the uri
    :param parameters: parameters to search through the uri.
    For example: OrderedDict({"virtualNetworks": "virtual_network_name"})
    """
    ret = []
    uri_values = uri.split("/")
    for parameter_key, parameter_value in parameters.items():
        if parameter_key in uri_values:
            uri_index = uri_values.index(parameter_key)
            # Convert parameters from camel case to snake case
            ret.append({parameter_value: uri_values[uri_index + 1]})
    return ret


def get_parameter_value_in_dict(hub, uri: str, parameters: OrderedDict) -> Dict:
    """
    Generate a list of pairs of uri parameter and its value. And convert uri parameters to expected key used in sls files
    For example /virtualNetworks/virtual-network-name/subnets/subnet-name} will output
    {"virtualNetworks": "virtual-network-name", "subnets": "subnet-name"}
    :param hub: The redistributed pop central hub. This is required in Idem, so while not used, must appear.
    :param uri: the uri
    :param parameters: parameters to search through the uri.
    For example: OrderedDict({"virtualNetworks": "virtual_network_name"})
    """
    ret = {}
    uri_values = uri.split("/")
    for parameter_key, parameter_value in parameters.items():
        if parameter_key in uri_values:
            uri_index = uri_values.index(parameter_key)
            # Convert parameters from camel case to snake case
            ret[parameter_value] = uri_values[uri_index + 1]
    return ret



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_key_vault_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    resource_id: str,
    vault_name: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        vault_name: Azure vault resource name.
        resource_id: Azure vault resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
      A Dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "vault_name": vault_name,
        "location": resource["location"],
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    if "properties" in resource:
        resource_translated["properties"] = resource["properties"]
    if "type" in resource:
        resource_translated["type"] = resource["type"]
    if "systemData" in resource:
        resource_translated["system_data"] = resource["systemData"]
    return resource_translated


def convert_present_to_key_vault(
    hub,
    location: str,
    sku: Dict[str, Any] = None,
    tags: Dict = None,
    tenant_id: str = None,
    enabled_for_deployment: bool = None,
    enabled_for_disk_encryption: bool = None,
    enabled_for_template_deployment: bool = None,
    enable_rbac_authorization: bool = None,
    public_network_access_enabled: str = None,
    soft_delete_retention_days: int = None,
    purge_protection_enabled: bool = None,
    access_policies: List[Dict[str, Any]] = None,
    network_acls: Dict[str, Any] = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location(str): Resource location. Changing this forces a new resource to be created.
        sku(Dict, Optional): The SKU of the key vault.
        tags(Dict, Optional): Resource tags.
        tenant_id(str, Optional): Tenant id of azure account.
        enabled_for_deployment(bool, Optional): Boolean flag to specify whether Azure Virtual Machines are permitted to retrieve certificates stored as secrets from the key vault.
        enabled_for_disk_encryption(bool, Optional): Boolean flag to specify whether Azure Disk Encryption is permitted to retrieve secrets from the vault and unwrap keys.
        enabled_for_template_deployment(bool, Optional):  Boolean flag to specify whether Azure Resource Manager is permitted to retrieve secrets from the key vault.
        enable_rbac_authorization(bool, Optional): Boolean flag to specify whether Azure Key Vault uses Role Based Access Control (RBAC) for authorization of data actions.
        public_network_access_enabled(str, Optional): Whether public network access is allowed for this Key Vault. Defaults to true
        soft_delete_retention_days(int, Optional): The number of days that items should be retained for once soft-deleted. This value can be between 7 and 90 (the default) days.
        purge_protection_enabled(bool, Optional): Is Purge Protection enabled for this Key Vault?
        access_policies(List[Dict[str, Any]], Optional): Key vault access policies.
        network_acls(Dict[str, Any], Optional): Key vault Network Acl.
    Returns:
        A Dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": {}}
    if tags is not None:
        payload["tags"] = tags
    payload["properties"]["sku"] = {
        "name": sku.get("name") if sku else "Standard",
        "family": sku.get("family") if sku else "A",
    }
    payload["properties"]["tenantId"] = tenant_id
    if enabled_for_deployment is not None:
        payload["properties"]["enabledForDeployment"] = enabled_for_deployment
    if enabled_for_disk_encryption is not None:
        payload["properties"]["enabledForDiskEncryption"] = enabled_for_disk_encryption
    if enabled_for_template_deployment is not None:
        payload["properties"][
            "enabledForTemplateDeployment"
        ] = enabled_for_template_deployment
    if public_network_access_enabled is not None:
        payload["properties"]["publicNetworkAccess"] = public_network_access_enabled
    if enable_rbac_authorization is not None:
        payload["properties"]["enableRbacAuthorization"] = enable_rbac_authorization
    if soft_delete_retention_days is not None:
        payload["properties"]["softDeleteRetentionInDays"] = soft_delete_retention_days
    if purge_protection_enabled is not None:
        payload["properties"]["enablePurgeProtection"] = purge_protection_enabled
    payload["properties"]["accessPolicies"] = (
        convert_present_to_access_policies(access_policies) if access_policies else []
    )
    if network_acls is not None:
        payload["properties"]["networkAcls"] = convert_present_to_network_acl(
            network_acls
        )
    return payload


def convert_present_to_access_policies(access_policies: List[Dict[str, Any]]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        access_policies(List[Dict[str, Any]]) : List of access policies for key vault.

    Returns:
        List of access policies for key vault payload List[Dict[str, Any]] in the format of an Azure PUT operation payload.
    """
    access_policies_list: List = []
    for access_policy in access_policies:
        access_policies_payload = {
            "tenantId": access_policy.get("tenant_id"),
            "objectId": access_policy.get("object_id"),
        }
        access_policies_payload["permissions"] = {}
        if access_policy.get("key_permissions") is not None:
            access_policies_payload["permissions"]["keys"] = access_policy.get(
                "key_permissions"
            )
        if access_policy.get("secret_permissions") is not None:
            access_policies_payload["permissions"]["secrets"] = access_policy.get(
                "secret_permissions"
            )
        if access_policy.get("certificate_permissions") is not None:
            access_policies_payload["permissions"]["certificates"] = access_policy.get(
                "certificate_permissions"
            )
        if access_policy.get("storage_permissions") is not None:
            access_policies_payload["permissions"]["storage"] = access_policy.get(
                "storage_permissions"
            )
        access_policies_list.append(access_policies_payload)
    return access_policies_list


def convert_present_to_network_acl(network_acls: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        network_acls(Dict(str, Any)) : Specifies information about the Network Acl for key vault.

    Returns:
        Network Acl Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    network_acls_payload = {
        "bypass": network_acls.get("bypass"),
        "defaultAction": network_acls.get("default_action"),
        "ipRules": network_acls.get("ip_rules"),
        "virtualNetworkRules": network_acls.get("virtual_network_subnet_ids"),
    }
    return network_acls_payload


def update_key_vault_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
    PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
        this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
        do an update.

    Returns:
        A result Dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    if need_update:
        result["ret"] = new_payload
    return result



RAW_TO_PRESENT_MAP = {
    "subscription_id": "subscriptionId",
    "resource_group_name": "resourceGroupName",
    "virtual_network_name": "virtualNetworkName",
    "virtual_network_peering_name": "virtualNetworkPeeringName",
    "allow_virtual_network_access": "properties.allowVirtualNetworkAccess",
    "allow_forwarded_traffic": "properties.allowForwardedTraffic",
    "allow_gateway_transit": "properties.allowGatewayTransit",
    "use_remote_gateways": "properties.useRemoteGateways",
    "remote_virtual_network": {"id": "properties.remoteVirtualNetwork.id"},
    "do_not_verify_remote_gateways": "properties.doNotVerifyRemoteGateways",
}


PRESENT_TO_RAW_MAP = {
    "subscriptionId": "subscription_id",
    "resourceGroupName": "resource_group_name",
    "virtualNetworkName": "virtual_network_name",
    "virtualNetworkPeeringName": "virtual_network_peering_name",
    "properties": {
        "allowVirtualNetworkAccess": "allow_virtual_network_access",
        "allowForwardedTraffic": "allow_forwarded_traffic",
        "allowGatewayTransit": "allow_gateway_transit",
        "useRemoteGateways": "use_remote_gateways",
        "remoteVirtualNetwork": {"id": "remote_virtual_network.id"},
        "doNotVerifyRemoteGateways": "do_not_verify_remote_gateways",
    },
}


def convert_raw_to_present_state(hub, raw_state):
    return hub.tool.azure.generic.convert_state_format(
        raw_state,
        RAW_TO_PRESENT_MAP,
    )


def convert_present_to_raw_state(hub, present_state):
    return hub.tool.azure.generic.convert_state_format(
        present_state,
        PRESENT_TO_RAW_MAP,
    )



import copy
from typing import Any
from typing import Dict

from dict_tools.differ import deep_diff


async def run_present(
    hub,
    ctx,
    resource_idem_path: str,
    api_version: str,
    name: str,
    resource_id: str,
    resource_id_format: str,
    path_properties: Dict,
    query_properties: Dict,
    body_properties: Dict,
) -> Dict[str, Any]:
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if not path_properties.get("subscription_id"):
        path_properties = copy.deepcopy(path_properties)
        path_properties["subscription_id"] = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = resource_id_format.format(**path_properties)

    get_method = getattr(hub.exec, f"{resource_idem_path}.get")
    response_get = await get_method(ctx, resource_id=resource_id)

    if response_get["result"]:
        path_properties = {k: v for k, v in path_properties.items() if v is not None}
        query_properties = {k: v for k, v in query_properties.items() if v is not None}
        body_properties = {k: v for k, v in body_properties.items() if v is not None}

        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_id": resource_id,
                        **path_properties,
                        **query_properties,
                        **body_properties,
                    },
                )

                result["comment"].append(
                    hub.tool.azure.comment_utils.would_create_comment(
                        resource_idem_path, name
                    )
                )
                return result
            else:
                # PUT operation to create a resource
                raw_body_payload = getattr(
                    hub.tool, f"{resource_idem_path}.convert_present_to_raw_state"
                )(body_properties)

                request_result = await hub.tool.azure.generic.run_put_request(
                    ctx,
                    resource_id,
                    path_properties,
                    query_properties,
                    api_version,
                    raw_body_payload,
                    resource_idem_path,
                    name,
                    "create",
                )
                result["result"] = request_result["result"]
                result["comment"] += request_result["comment"]
                result["new_state"] = request_result["new_state"]
                return result

        else:
            result["old_state"] = response_get["ret"]
            result["old_state"]["name"] = name
            present_update_payload = get_update_payload(
                response_get["ret"], body_properties, path_properties, query_properties
            )
            if present_update_payload:
                if ctx.get("test", False):
                    # Return a proposed state by Idem state --test
                    result[
                        "new_state"
                    ] = hub.tool.azure.test_state_utils.generate_test_state(
                        enforced_state={},
                        desired_state={
                            "name": name,
                            "resource_id": resource_id,
                            **path_properties,
                            **query_properties,
                            **body_properties,
                        },
                    )

                    result["comment"].append(
                        hub.tool.azure.comment_utils.would_update_comment(
                            resource_idem_path, name
                        )
                    )
                    return result

                # PUT operation to update a resource
                update_payload = getattr(
                    hub.tool, f"{resource_idem_path}.convert_present_to_raw_state"
                )(present_update_payload)
                request_result = await hub.tool.azure.generic.run_put_request(
                    ctx,
                    resource_id,
                    path_properties,
                    query_properties,
                    api_version,
                    update_payload,
                    resource_idem_path,
                    name,
                    "update",
                )
                result["result"] = request_result["result"]
                result["comment"] += request_result["comment"]
                result["new_state"] = request_result["new_state"]
            else:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                        resource_idem_path, name
                    )
                )
                return result

            return result

    else:
        hub.log.debug(
            f"{hub.tool.azure.comment_utils.could_not_get_comment(resource_idem_path, name)} {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def run_absent(
    hub,
    ctx,
    resource_idem_path: str,
    api_version: str,
    name: str,
    resource_id_format: str,
    resource_id: str,
    path_properties: Dict,
    query_properties: Dict,
) -> Dict[str, Any]:
    result = dict(name=name, result=True, comment=[], old_state=None, new_state=None)

    if not path_properties.get("subscription_id"):
        path_properties = copy.deepcopy(path_properties)
        path_properties["subscription_id"] = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = resource_id_format.format(**path_properties)

    get_method = getattr(hub.exec, f"{resource_idem_path}.get")
    response_get = await get_method(ctx, resource_id=resource_id)
    if response_get["result"]:
        if response_get["ret"]:
            result["old_state"] = response_get["ret"]
            result["old_state"]["name"] = name
            if ctx.get("test", False):
                result["comment"].append(
                    hub.tool.azure.comment_utils.would_delete_comment(
                        resource_idem_path, name
                    )
                )
                return result
            query_properties_url = "&".join(
                [
                    f"{prop_name}={prop_val}"
                    for prop_name, prop_val in query_properties.items()
                ]
            )
            url = f"{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}"
            if query_properties_url:
                url = f"{url}&{query_properties_url}"
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=url,
                success_codes=[200, 202, 204],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"{hub.tool.azure.comment_utils.could_not_delete_comment(resource_idem_path, name)} {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_delete["comment"], response_delete["ret"]]
                return result

            result["comment"].append(
                hub.tool.azure.comment_utils.delete_comment(resource_idem_path, name)
            )
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                hub.tool.azure.comment_utils.already_absent_comment(
                    resource_idem_path, name
                )
            )
            return result
    else:
        hub.log.debug(
            f"{hub.tool.azure.comment_utils.could_not_get_comment(resource_idem_path, name)}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def run_put_request(
    hub,
    ctx,
    resource_id: str,
    path_properties: str,
    query_properties: Dict,
    api_version: str,
    raw_payload: Dict,
    resource_idem_path: str,
    name: str,
    mode: str,
) -> Dict[str, Any]:
    result = {"comment": [], "result": True, "new_state": None}

    fail_comment = ""
    success_comment = ""
    if mode == "create":
        fail_comment = hub.tool.azure.comment_utils.could_not_create_comment(
            resource_idem_path, name
        )
        success_comment = hub.tool.azure.comment_utils.create_comment(
            resource_idem_path, name
        )
    elif mode == "update":
        fail_comment = hub.tool.azure.comment_utils.could_not_update_comment(
            resource_idem_path, name
        )
        success_comment = hub.tool.azure.comment_utils.update_comment(
            resource_idem_path, name
        )

    query_properties_url = "&".join(
        [f"{prop_name}={prop_val}" for prop_name, prop_val in query_properties.items()]
    )
    url = f"{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}"
    if query_properties_url:
        url = f"{url}&{query_properties_url}"
    response = await hub.exec.request.json.put(
        ctx,
        url=url,
        success_codes=[200, 201, 202],
        json=raw_payload,
    )

    if not response["result"]:
        hub.log.debug(f"{fail_comment} {response['comment']} {response['ret']}")
        result["comment"] = [response["comment"], response["ret"]]
        result["result"] = False
        return result

    if response["status"] == 202:
        # TODO: Use reconciler mechanism to wait for operation to complete
        response = await hub.tool.azure.polling.await_operation(
            ctx,
            initial_response=response,
            resource_url=f"{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}",
        )
    elif not response["result"] and response["status"] != 202:
        hub.log.debug(f"{fail_comment} {response['comment']} {response['ret']}")
        result["comment"] = [response["comment"], response["ret"]]
        result["result"] = False
        return result

    result["new_state"] = getattr(
        hub.tool, f"{resource_idem_path}.convert_raw_to_present_state"
    )(response["ret"])
    result["new_state"].update(path_properties)
    result["new_state"].update(query_properties)
    result["new_state"].update({"name": name, "resource_id": resource_id})
    result["comment"].append(success_comment)
    return result


def get_update_payload(
    old_state: Dict,
    new_body_properties: Dict,
    path_properties: Dict,
    query_properties: Dict,
) -> Dict[str, Any]:
    # Exclude the path and query properties from the diff
    old_body_properties = copy.deepcopy(old_state)
    for prop_name in old_state:
        if prop_name in path_properties or prop_name in query_properties:
            old_body_properties.pop(prop_name, None)

    has_change = False
    changed_props_payload = {}
    for new_body_prop_name, new_body_prop_value in new_body_properties.items():
        if new_body_prop_value is not None:
            changed_props_payload[new_body_prop_name] = new_body_prop_value
            if deep_diff(
                {new_body_prop_name: old_body_properties.get(new_body_prop_name)},
                {new_body_prop_name: new_body_prop_value},
            ):
                has_change = True

    if has_change:
        changed_props_payload = copy.deepcopy(new_body_properties)
        for old_body_prop_name, old_body_prop_value in old_body_properties.items():
            if old_body_prop_name not in changed_props_payload:
                changed_props_payload[old_body_prop_name] = old_body_prop_value

        return changed_props_payload

    return {}


def convert_state_format(
    hub, state: Any, formatting_map: Any, list_items_indexes: Dict = None
) -> Any:
    """
    Converts a state from one format to another, e.g. present (snake case) to raw (camel case) format or vice versa.
    The conversion can change the property names and the structure of the resulting state, e.g. flatten
    deeply nested properties. Works with nested lists and dicts.

    Args:
        hub: The redistributed pop central hub.
        state: the current state of a resource.
        formatting_map: a mapping from one format to another. E.g.:
            "virtualNetworkPeerings": [
                {
                    "$list_refs": ["virtual_network_peerings"],
                    "name": "virtual_network_peerings.name",
                    "type": "virtual_network_peerings.type",
                    "properties": {
                        "allowVirtualNetworkAccess": "virtual_network_peerings.allow_virtual_network_access",
                        "allowForwardedTraffic": "virtual_network_peerings.allow_forwarded_traffic",
                        "allowGatewayTransit": "virtual_network_peerings.allow_gateway_transit",
                        "useRemoteGateways": "virtual_network_peerings.use_remote_gateways",
                        "remoteVirtualNetwork": "virtual_network_peerings.remote_virtual_network",
                        "remoteAddressSpace": {
                            "addressPrefixes": "virtual_network_peerings.remote_address_space.address_prefixes"
                        },
                        "remoteVirtualNetworkAddressSpace": {
                            "addressPrefixes": "virtual_network_peerings.remote_virtual_network_address_space.address_prefixes"
                        },
                        "remoteBgpCommunities": {
                            "virtualNetworkCommunity": "virtual_network_peerings.remote_bgp_communities.virtual_network_community"
                        },
                        "peeringState": "virtual_network_peerings.peering_state",
                        "peeringSyncLevel": "virtual_network_peerings.peering_sync_level",
                        "doNotVerifyRemoteGateways": "virtual_network_peerings.do_not_verify_remote_gateways",
                    },
                }
            ]
            $list_refs is a special reserved property in the mapping. Use it to specify which lists from the original
            state should be tracked at the current nesting level
        list_items_indexes: ignore, used internally by this recursive method.
            This property keeps track of the nested lists indexes

    Returns:
        The state in the new format
    """
    formatted_state = {}

    if isinstance(formatting_map, dict):
        for formatted_name, formatting_map_for_item in formatting_map.items():
            if formatted_name == "$list_refs":
                continue

            converted_value = hub.tool.azure.generic.convert_state_format(
                state, formatting_map_for_item, list_items_indexes
            )
            if (
                isinstance(formatting_map_for_item, str) and converted_value is not None
            ) or (
                isinstance(formatting_map_for_item, (dict, list)) and converted_value
            ):
                formatted_state[formatted_name] = converted_value
        return formatted_state
    elif isinstance(formatting_map, list):
        ref_lists_names = formatting_map[0]["$list_refs"]
        if not isinstance(ref_lists_names, list):
            ref_lists_names = [ref_lists_names]
        ref_lists = [
            _get_element_from_state(state, ref_list_name, list_items_indexes)
            for ref_list_name in ref_lists_names
        ]
        longest_ref_list = max(ref_lists, key=len)
        formatted_list = []
        for i, item in enumerate(longest_ref_list):
            if list_items_indexes is None:
                list_items_indexes = {}
            for ref_list_name in ref_lists_names:
                list_items_indexes[ref_list_name] = i
            formatted_list.append(
                hub.tool.azure.generic.convert_state_format(
                    state, formatting_map[0], list_items_indexes
                )
            )

        if list_items_indexes:
            for ref_list_name in ref_lists_names:
                list_items_indexes.pop(ref_list_name, None)

        return formatted_list
    elif isinstance(formatting_map, str):
        return _get_element_from_state(state, formatting_map, list_items_indexes)

    return formatted_state


def _get_element_from_state(
    state: Any, path: str, list_item_indexes: Dict = None
) -> Any:
    path_segments = path.split(".")
    partial_path = None
    element = state
    for path_segment in path_segments:
        if element.get(path_segment) is None:
            return None
        element = element[path_segment]
        partial_path = (
            f"{partial_path}.{path_segment}" if partial_path else path_segment
        )
        if list_item_indexes and partial_path in list_item_indexes:
            i = list_item_indexes[partial_path]
            if i < len(element):
                element = element[i]
            else:
                return None

    return element



RESOURCES_WITH_PRESENT_WRAPPER = [
    "azure.sql_database.databases",
]


async def call_present(hub, ctx):
    r"""Wrapper for present function."""

    name = ctx.kwargs.get("name", None)
    state_ctx = ctx.kwargs.get("ctx") or ctx.args[1]
    assert state_ctx, f"state context is missing: {state_ctx}"

    azure_service_resource_type = state_ctx.get("tag").split("_|")[0]

    # TODO: This needs to be removed once all resources follow the contract
    if azure_service_resource_type not in RESOURCES_WITH_PRESENT_WRAPPER:
        return await ctx.func(*ctx.args, **ctx.kwargs)

    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    ctx.kwargs[
        "subscription_id"
    ] = hub.tool.azure.resource_utils.get_subscription_id_from_account(
        state_ctx, ctx.kwargs.get("subscription_id")
    )

    service_resource_type = azure_service_resource_type.replace("azure.", "")

    hub_ref_exec = hub.exec.azure
    for resource_path_segment in service_resource_type.split("."):
        hub_ref_exec = hub_ref_exec[resource_path_segment]

    resource_id = ctx.kwargs.get("resource_id")
    local_params = {**ctx.kwargs}

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    if resource_id:
        response_get = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not response_get["result"] or (
            not response_get["ret"] and get_resource_only_with_resource_id
        ):
            result["result"] = False
            result["comment"] += response_get["comment"]
            return result

        result["old_state"] = response_get["ret"]
    elif not get_resource_only_with_resource_id:
        resource_id = hub.tool.azure.resource_utils.construct_resource_id(
            service_resource_type, local_params
        )

        if not resource_id:
            result["result"] = False
            result["comment"].append(
                f"Could not construct resource ID of {azure_service_resource_type} from input arguments."
            )
            return result

        response_get = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not response_get["result"]:
            result["result"] = False
            result["comment"] += response_get["comment"]
            return result

        if response_get["ret"]:
            result["old_state"] = response_get["ret"]

    state_ctx["wrapper_result"] = result
    state_ctx["computed"] = {
        "resource_url": hub.tool.azure.resource_utils.construct_resource_url(
            state_ctx, service_resource_type, local_params
        ),
    }
    return await ctx.func(*ctx.args, **{**ctx.kwargs, "resource_id": resource_id})



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_raw_network_security_groups_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_id: str,
    resource_group_name: str,
    network_security_group_name: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Converts raw network_security_group api response into present function format

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        network_security_group_name: Azure Network Security group name.
        resource_id: Azure Policy Definition resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "network_security_group_name": network_security_group_name,
        "location": resource["location"],
    }
    if resource.get("tags"):
        translated_resource["tags"] = resource["tags"]
    if resource["properties"].get("securityRules"):
        translated_resource[
            "security_rules"
        ] = convert_raw_to_present_nsg_security_rule(
            hub, resource["properties"]["securityRules"]
        )
    return translated_resource


def convert_present_to_raw_network_security_groups(
    hub,
    location: str,
    security_rules: List = None,
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location: Resource location.
        security_rules: List of Security rules
        tags: Resource tags.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": {}}
    if tags is not None:
        payload["tags"] = tags
    if security_rules is not None:
        payload["properties"][
            "securityRules"
        ] = convert_present_to_raw_nsg_security_rule(hub, security_rules)
    return payload


def convert_present_to_raw_nsg_security_rule(
    hub, security_rules: List[Dict[str, Any]]
) -> List:
    """
    Converts security rule from present format to raw
    Args:
        security_rules: Raw security rules

    Returns:
        List
    """
    raw_security_rules = []
    for security_rule in security_rules:
        payload = {
            "name": security_rule["name"],
            "properties": {
                "protocol": security_rule["protocol"],
                "access": security_rule["access"],
                "priority": security_rule["priority"],
                "direction": security_rule["direction"],
            },
        }
        if security_rule.get("source_port_range"):
            payload["properties"]["sourcePortRange"] = security_rule[
                "source_port_range"
            ]
        if security_rule.get("source_port_ranges"):
            payload["properties"]["sourcePortRanges"] = security_rule[
                "source_port_ranges"
            ]
        if security_rule.get("destination_port_range"):
            payload["properties"]["destinationPortRange"] = security_rule[
                "destination_port_range"
            ]
        if security_rule.get("destination_port_ranges"):
            payload["properties"]["destinationPortRanges"] = security_rule[
                "destination_port_ranges"
            ]

        if security_rule.get("source_address_prefix"):
            payload["properties"]["sourceAddressPrefix"] = security_rule[
                "source_address_prefix"
            ]
        if security_rule.get("source_address_prefixes"):
            payload["properties"]["sourceAddressPrefixes"] = security_rule[
                "source_address_prefixes"
            ]
        if security_rule.get("destination_address_prefix"):
            payload["properties"]["destinationAddressPrefix"] = security_rule[
                "destination_address_prefix"
            ]
        if security_rule.get("destination_address_prefixes"):
            payload["properties"]["destinationAddressPrefixes"] = security_rule[
                "destination_address_prefixes"
            ]

        raw_security_rules.append(payload)
    return raw_security_rules


def convert_raw_to_present_nsg_security_rule(
    hub, security_rules: List[Dict[str, Any]]
) -> List:
    """
    Converts raw security rule to present function input format
    Args:
        security_rules: Raw security rules

    Returns:
        List
    """
    present_security_rules = []
    for security_rule in security_rules:
        payload = {
            "name": security_rule["name"],
            "priority": security_rule["properties"]["priority"],
            "direction": security_rule["properties"]["direction"],
            "access": security_rule["properties"]["access"],
            "protocol": security_rule["properties"]["protocol"],
        }
        if security_rule["properties"].get("sourcePortRange"):
            payload["source_port_range"] = security_rule["properties"][
                "sourcePortRange"
            ]
        if security_rule["properties"].get("sourcePortRanges"):
            payload["source_port_ranges"] = security_rule["properties"][
                "sourcePortRanges"
            ]
        if security_rule["properties"].get("destinationPortRange"):
            payload["destination_port_range"] = security_rule["properties"][
                "destinationPortRange"
            ]
        if security_rule["properties"].get("destinationPortRanges"):
            payload["destination_port_ranges"] = security_rule["properties"][
                "destinationPortRanges"
            ]

        if security_rule["properties"].get("sourceAddressPrefix"):
            payload["source_address_prefix"] = security_rule["properties"][
                "sourceAddressPrefix"
            ]
        if security_rule["properties"].get("sourceAddressPrefixes"):
            payload["source_address_prefixes"] = security_rule["properties"][
                "sourceAddressPrefixes"
            ]
        if security_rule["properties"].get("destinationAddressPrefix"):
            payload["destination_address_prefix"] = security_rule["properties"][
                "destinationAddressPrefix"
            ]
        if security_rule["properties"].get("destinationAddressPrefixes"):
            payload["destination_address_prefixes"] = security_rule["properties"][
                "destinationAddressPrefixes"
            ]
        present_security_rules.append(payload)
    return present_security_rules


def update_network_security_groups_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    new_payload["properties"] = {}
    existing_properties = existing_payload["properties"]

    if new_values.get("security_rules") is not None:
        existing_security_rules = existing_properties.get("securityRules")
        existing_security_rules_required_payload = (
            convert_raw_to_present_nsg_security_rule(
                hub, security_rules=existing_security_rules
            )
        )
        if diff_nsg_security_rules(
            hub,
            new_values.get("security_rules"),
            existing_security_rules_required_payload,
            "name",
        ):
            new_payload["properties"][
                "securityRules"
            ] = convert_present_to_raw_nsg_security_rule(
                hub, security_rules=new_values.get("security_rules")
            )
            need_update = True

    if need_update:
        result["ret"] = new_payload
    return result


def diff_nsg_security_rules(
    hub,
    new_values: List[Dict[str, any]],
    old_values: List[Dict[str, any]],
    sorting_key: str,
):
    """
    Compares security rule variable and returns true if update is required.

    Args:
        new_values(List): List of new security rules
        old_values(List): List of new security rules
        sorting_key(str): Sorting key value

    Returns:
        bool
    """
    for entry in new_values:
        for key, value in entry.items():
            if isinstance(entry[key], list):
                entry[key].sort()

    for entry in old_values:
        for key, value in entry.items():
            if isinstance(entry[key], list):
                entry[key].sort()
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        old_values, key=lambda x: x[sorting_key]
    )



"""State module for managing Network interface."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_interface_name: str,
    location: str,
    ip_configurations: List[
        make_dataclass(
            "IpConfigurationsSet",
            [
                ("name", str, field(default=None)),
                ("private_ip_address_allocation", str, field(default=None)),
                ("subnet_id", str, field(default=None)),
                ("private_ip_address_version", str, field(default=None)),
                ("private_ip_address", str, field(default=None)),
                ("primary", bool, field(default=False)),
            ],
        )
    ],
    tags: Dict = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Network Interfaces.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_interface_name(str): The name of the network interface.
        location(str): Resource location. This field can not be updated.
        ip_configurations(list[dict[str, Any]]): A list of IPConfigurations of the network interface. Each ip configuration supports fields:

            * name(str):
                The name of the resource that is unique within a resource group.
            * private_ip_address_allocation(str):
                The allocation method used for the Private IP Address. Possible values are Dynamic and Static.
                Azure does not assign a Dynamic IP Address until the Network Interface is attached to a running Virtual Machine(or other resource).
            * subnet_id(str):
                Resource ID of the Subnet bound to the IP configuration. The name of the resource that is unique within a resource group.
                This is required when private_ip_address_version is set to IPv4.
            * private_ip_address_version(str):
                The specific IP configuration is IPv4 or IPv6. Default is IPv4.
            * private_ip_address(str):
                The Static IP Address which should be used. When private_ip_address_allocation is set to Static, private_ip_address can be configured.
            * primary(bool):
                To check if this is the primary IP Configuration. Must be true for the first ip_configuration when
                multiple are specified. Defaults to false. Primary attribute must be true for the first ip_configuration
                when multiple are specified. Defaults to false.
        tags(dict[str, str], Optional): Resource tags.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Network Interface resource id on Azure.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            my-nic:
              azure.network.network_interfaces.present:
                - name: my-nic
                - resource_group_name: my-rg
                - network_interface_name: my-nic
                - location: southindia
                - subscription_id: my-subscription
                - ip_configurations:
                  - name: my-ipc
                    private_ip_address_allocation: Static
                    subnet_id: subnet_name
                    private_ip_address_version: IPv4
                    private_ip_address: 10.0.0.24
                    primary: true
                - tags:
                    my-tag-key: my-tag-value

    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = (
            f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}"
            f"/providers/Microsoft.Network/networkInterfaces/{network_interface_name}"
        )
    response_get = await hub.exec.azure.network.network_interfaces.get(
        ctx, resource_id=resource_id, raw=True
    )

    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "subscription_id": subscription_id,
                        "network_interface_name": network_interface_name,
                        "location": location,
                        "ip_configurations": ip_configurations,
                        "tags": tags,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.network.network_interfaces '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.network.network_interfaces.convert_present_to_raw_network_interfaces(
                    location=location,
                    ip_configurations=ip_configurations,
                    tags=tags,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-08-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.network.network_interfaces"
                        f" '{name}' {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.network_interfaces.convert_raw_network_interfaces_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    subscription_id=subscription_id,
                    network_interface_name=network_interface_name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Created azure.network.network_interfaces '{name}'"
                )
                return result
        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.network.network_interfaces.convert_raw_network_interfaces_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                subscription_id=subscription_id,
                network_interface_name=network_interface_name,
                resource_id=resource_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.network.network_interfaces.update_network_interfaces_payload(
                existing_resource,
                {
                    "ip_configurations": ip_configurations,
                    "tags": tags,
                },
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.network.network_interfaces '{name}' has no property need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.network.network_interfaces.convert_raw_network_interfaces_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        subscription_id=subscription_id,
                        network_interface_name=network_interface_name,
                        resource_id=resource_id,
                    )
                    result["comment"].append(
                        f"Would update azure.network.network_interfaces '{name}'"
                    )
                return result

                # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.network_interfaces '{name}' has no property need to be updated."
                )
                return result
            result["comment"] = result["comment"] + new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-08-01",
                success_codes=[200, 201],
                json=new_payload["ret"],
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.network.network_interfaces "
                    f"{response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.network_interfaces.convert_raw_network_interfaces_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                subscription_id=subscription_id,
                network_interface_name=network_interface_name,
                resource_id=resource_id,
            )
            if result["old_state"] == result["new_state"]:
                result["comment"].append(
                    f"azure.network.network_interfaces '{name}' has no property to be updated."
                )
                return result
            result["comment"].append(
                f"Updated azure.network.network_interfaces '{name}'"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.network_interfaces {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_interface_name: str,
    subscription_id: str = None,
) -> Dict:
    r"""Delete Network Interfaces.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_interface_name(str): The name of the network interface.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.network_interfaces.absent:
                - name: value
                - resource_group_name: value
                - network_interface_name: value
                - subscription_id: my-subscription
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = (
        f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}"
        f"/providers/Microsoft.Network/networkInterfaces/{network_interface_name}"
    )

    response_get = await hub.exec.azure.network.network_interfaces.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"]:
            result[
                "old_state"
            ] = hub.tool.azure.network.network_interfaces.convert_raw_network_interfaces_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                subscription_id=subscription_id,
                network_interface_name=network_interface_name,
                resource_id=resource_id,
            )
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.network.network_interfaces '{name}'"
                )
                return result

            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-08-01",
                success_codes=[200, 202],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.network.network_interfaces {response_delete['comment']}"
                    f" {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_delete["comment"], response_delete["ret"]]
                return result

            result["comment"].append(
                f"Deleted azure.network.network_interfaces '{name}'"
            )
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.network.network_interfaces '{name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.network_interfaces '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Network Interfaces under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.network_interfaces
    """
    result = {}
    ret_list = await hub.exec.azure.network.network_interfaces.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe network_interfaces {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.network_interfaces.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



import copy
from typing import Any
from typing import Dict


def convert_raw_log_analytics_workspace_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    workspace_name: str,
    resource_id: str,
    subscription_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        workspace_name: Azure Workspace resource name.
        resource_id: Azure Log Analytics Workspace resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "workspace_name": workspace_name,
        "location": resource["location"],
        "subscription_id": subscription_id,
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "retentionInDays": "retention_in_days",
            "publicNetworkAccessForIngestion": "public_network_access_for_ingestion",
            "publicNetworkAccessForQuery": "public_network_access_for_query",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties.get(parameter_raw)

        if properties.get("features") is not None:
            existing_features_required_payload = convert_raw_features_to_present(
                features=properties["features"]
            )
            resource_translated["features"] = existing_features_required_payload
        if properties.get("sku") is not None:
            existing_sku_required_payload = convert_raw_sku_to_present(
                sku=properties["sku"]
            )
            resource_translated["sku"] = existing_sku_required_payload
        if properties.get("workspaceCapping") is not None:
            existing_workspace_capping_required_payload = (
                convert_raw_workspace_capping_to_present(
                    workspace_capping=properties["workspaceCapping"]
                )
            )
            resource_translated[
                "workspace_capping"
            ] = existing_workspace_capping_required_payload
    return resource_translated


def convert_present_log_analytics_workspace_to_raw(
    hub,
    subscription_id: str,
    public_network_access_for_query: str = None,
    public_network_access_for_ingestion: str = None,
    retention_in_days: int = None,
    workspace_capping: Dict = None,
    features: Dict = None,
    sku: Dict = None,
    location: str = None,
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id
        public_network_access_for_query: The network access type for accessing Log Analytics query.
        public_network_access_for_ingestion: The network access type for accessing Log Analytics ingestion.
        retention_in_days: The workspace data retention in days. Allowed values are per pricing plan.
        workspace_capping(dict, optional): The daily volume cap for ingestion.
        features(list[dict], optional): Workspace features.
        sku(dict, optional): The SKU of the workspace.
        location: Resource location. This field can not be updated.
        tags: Resource tags.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": {}}
    if tags is not None:
        payload["tags"] = tags
    if retention_in_days is not None:
        payload["properties"]["retentionInDays"] = retention_in_days
    if public_network_access_for_ingestion is not None:
        payload["properties"][
            "publicNetworkAccessForIngestion"
        ] = public_network_access_for_ingestion
    if public_network_access_for_query is not None:
        payload["properties"][
            "publicNetworkAccessForQuery"
        ] = public_network_access_for_query
    if sku is not None:
        sku_payload = convert_present_sku_to_raw(sku)
        payload["properties"]["sku"] = sku_payload
    if features is not None:
        features_payload = convert_present_features_to_raw(features)
        payload["properties"]["features"] = features_payload
    if workspace_capping is not None:
        workspace_capping_payload = convert_present_workspace_capping_to_raw(
            workspace_capping
        )
        payload["properties"]["workspaceCapping"] = workspace_capping_payload
    return payload


def update_log_analytics_workspace_payload(
    hub,
    subscription_id: str,
    existing_payload: Dict[str, Any],
    new_values: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id required to update network security group for VN subnet
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    is_updated = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        is_updated = True
    existing_properties = existing_payload["properties"]
    if (new_values.get("retention_in_days") is not None) and (
        new_values["retention_in_days"] != existing_properties.get("retentionInDays")
    ):
        new_payload["properties"]["retentionInDays"] = new_values.get(
            "retention_in_days"
        )
        is_updated = True
    if (new_values.get("public_network_access_for_ingestion") is not None) and (
        new_values["public_network_access_for_ingestion"]
        != existing_properties.get("publicNetworkAccessForIngestion")
    ):
        new_payload["properties"]["publicNetworkAccessForIngestion"] = new_values.get(
            "public_network_access_for_ingestion"
        )
        is_updated = True
    if (new_values.get("public_network_access_for_query") is not None) and (
        new_values["public_network_access_for_query"]
        != existing_properties.get("publicNetworkAccessForQuery")
    ):
        new_payload["properties"]["publicNetworkAccessForQuery"] = new_values.get(
            "public_network_access_for_query"
        )
        is_updated = True
    if (new_values.get("sku") is not None) and (
        compare_accurate_object(
            new_values.get("sku"),
            convert_raw_sku_to_present(existing_properties["sku"]),
        )
    ):
        new_payload["properties"]["sku"] = convert_present_sku_to_raw(new_values["sku"])
        is_updated = True
    if (new_values.get("features") is not None) and compare_accurate_object(
        new_values.get("features"),
        convert_raw_features_to_present(existing_properties["features"]),
    ):
        new_payload["properties"]["features"] = convert_present_features_to_raw(
            new_values.get("features")
        )
        is_updated = True
    if (new_values.get("workspace_capping") is not None) and compare_accurate_object(
        new_values.get("features"),
        convert_raw_workspace_capping_to_present(
            existing_properties.get("workspaceCapping")
        ),
    ):
        new_payload["properties"][
            "workspaceCapping"
        ] = convert_present_workspace_capping_to_raw(
            new_values.get("workspace_capping")
        )
        is_updated = True
    if is_updated:
        result["ret"] = new_payload
    return result


def convert_present_features_to_raw(features: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        features(Dict[str, Any]) : Feature for Log Analytics creation

    Returns:
        Features payload : Dict[str,any] in the format of an Azure PUT operation payload.
    """
    features_payload = {}
    if features.get("enable_log_access_using_only_resource_permissions") is not None:
        features_payload["enableLogAccessUsingOnlyResourcePermissions"] = features.get(
            "enable_log_access_using_only_resource_permissions"
        )
    if features.get("legacy") is not None:
        features_payload["legacy"] = features.get("legacy")
    if features.get("search_version") is not None:
        features_payload["searchVersion"] = features.get("search_version")
    if features.get("clusterResourceId") is not None:
        features_payload["cluster_resource_id"] = features.get("cluster_resource_id")
    if features.get("enable_data_export") is not None:
        features_payload["enableDataExport"] = features.get("enable_data_export")
    if features.get("disable_local_auth") is not None:
        features_payload["disableLocalAuth"] = features.get("disable_local_auth")
    if features.get("immediate_purge_data_on_30_days") is not None:
        features_payload["immediatePurgeDataOn30Days"] = features.get(
            "immediate_purge_data_on_30_days"
        )

    return features_payload


def convert_raw_features_to_present(features: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        features(Dict, optional): Features payload in a log analytics resource

    Returns:
         Log analytics payload that contains the parameters that match respective present function's input format.
    """
    features_payload = {}
    if features.get("enableLogAccessUsingOnlyResourcePermissions") is not None:
        features_payload[
            "enable_log_access_using_only_resource_permissions"
        ] = features.get("enableLogAccessUsingOnlyResourcePermissions")
    if features.get("legacy") is not None:
        features_payload["legacy"] = features.get("legacy")
    if features.get("searchVersion") is not None:
        features_payload["search_version"] = features.get("searchVersion")
    if features.get("clusterResourceId") is not None:
        features_payload["cluster_resource_id"] = features.get("clusterResourceId")
    if features.get("enableDataExport") is not None:
        features_payload["enable_data_export"] = features.get("enableDataExport")
    if features.get("disableLocalAuth") is not None:
        features_payload["disable_local_auth"] = features.get("disableLocalAuth")
    if features.get("immediatePurgeDataOn30Days") is not None:
        features_payload["immediate_purge_data_on_30_days"] = features.get(
            "immediatePurgeDataOn30Days"
        )

    return features_payload


def convert_present_sku_to_raw(sku: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        sku(Dict[str, Any]) : Feature for Log Analytics creation

    Returns:
        sku payload : Dict[str,any] in the format of an Azure PUT operation payload.
    """
    sku_payload = {}
    if sku.get("capacity_reservation_level") is not None:
        sku_payload["capacityReservationLevel"] = sku.get("capacity_reservation_level")
    if sku.get("last_sku_update") is not None:
        sku_payload["lastSkuUpdate"] = sku.get("last_sku_update")
    if sku.get("name") is not None:
        sku_payload["name"] = sku.get("name")

    return sku_payload


def convert_raw_sku_to_present(sku: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        sku(Dict, optional): Features payload in a log analytics resource

    Returns:
         Log analytics payload that contains the parameters that match respective present function's input format.
    """
    sku_payload = {}
    if sku.get("capacityReservationLevel") is not None:
        sku_payload["capacity_reservation_level"] = sku.get("capacityReservationLevel")
    if sku.get("lastSkuUpdate") is not None:
        sku_payload["last_sku_update"] = sku.get("lastSkuUpdate")
    if sku.get("name") is not None:
        sku_payload["name"] = sku.get("name")

    return sku_payload


def convert_present_workspace_capping_to_raw(workspace_capping: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        workspace_capping(Dict[str, Any]) : Feature for Log Analytics creation

    Returns:
        Features payload : Dict[str,any] in the format of an Azure PUT operation payload.
    """
    workspace_capping_payload = {}
    if workspace_capping.get("daily_quota_gb") is not None:
        workspace_capping_payload["dailyQuotaGb"] = workspace_capping.get(
            "daily_quota_gb"
        )
    if workspace_capping.get("data_ingestion_status") is not None:
        workspace_capping_payload["dataIngestionStatus"] = workspace_capping.get(
            "data_ingestion_status"
        )
    if workspace_capping.get("quota_next_reset_time") is not None:
        workspace_capping_payload["quotaNextResetTime"] = workspace_capping.get(
            "quota_next_reset_time"
        )

    return workspace_capping_payload


def convert_raw_workspace_capping_to_present(workspace_capping: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        workspace_capping(Dict, optional): Features payload in a log analytics resource

    Returns:
         Log analytics payload that contains the parameters that match respective present function's input format.
    """
    workspace_capping_payload = {}
    if workspace_capping.get("dailyQuotaGb") is not None:
        workspace_capping_payload["daily_quota_gb"] = workspace_capping.get(
            "dailyQuotaGb"
        )
    if workspace_capping.get("dataIngestionStatus") is not None:
        workspace_capping_payload["data_ingestion_status"] = workspace_capping.get(
            "dataIngestionStatus"
        )
    if workspace_capping.get("quotaNextResetTime") is not None:
        workspace_capping_payload["quota_next_reset_time"] = workspace_capping.get(
            "quotaNextResetTime"
        )

    return workspace_capping_payload


def compare_accurate_object(
    new_object: Dict[str, Any], existing_object: Dict[str, Any]
):
    is_updated = False
    if len(existing_object) != len(new_object):
        return True
    for existing_key, existing_value in existing_object.items():
        new_value = new_object.get(existing_key)
        if new_value != existing_value:
            is_updated = True

    return is_updated



from typing import Any
from typing import Dict


def fill_properties_original(translated_properties, source_properties):
    # default_encryption_scope
    if "defaultEncryptionScope" in source_properties:
        translated_properties["default_encryption_scope"] = source_properties[
            "defaultEncryptionScope"
        ]

    # deny_encryption_scope_override
    if "denyEncryptionScopeOverride" in source_properties:
        translated_properties["deny_encryption_scope_override"] = source_properties[
            "denyEncryptionScopeOverride"
        ]

    # has_immutability_policy
    if "hasImmutabilityPolicy" in source_properties:
        translated_properties["has_immutability_policy"] = source_properties[
            "hasImmutabilityPolicy"
        ]

    # immutable_storage_with_versioning, ??
    if "immutable_storage_with_versioning" in source_properties:
        translated_properties["immutable_storage_with_versioning"] = source_properties[
            "default_encryption_scope"
        ]

    # metadata,
    if "metadata" in source_properties:
        translated_properties["metadata"] = source_properties["metadata"]

    # public_access
    if "publicAccess" in source_properties:
        translated_properties["public_access"] = source_properties["publicAccess"]

    return translated_properties


def fill_properties_translated(translated_properties, source_properties):
    # default_encryption_scope
    if "default_encryption_scope" in source_properties:
        translated_properties["default_encryption_scope"] = source_properties[
            "default_encryption_scope"
        ]

    # deny_encryption_scope_override
    if "deny_encryption_scope_override" in source_properties:
        translated_properties["deny_encryption_scope_override"] = source_properties[
            "deny_encryption_scope_override"
        ]

    # has_immutability_policy
    if "has_immutability_policy" in source_properties:
        translated_properties["has_immutability_policy"] = source_properties[
            "has_immutability_policy"
        ]

    # immutable_storage_with_versioning, ??
    if "immutable_storage_with_versioning" in source_properties:
        translated_properties["immutable_storage_with_versioning"] = source_properties[
            "immutable_storage_with_versioning"
        ]

    # metadata,
    if "metadata" in source_properties:
        translated_properties["metadata"] = source_properties["metadata"]

    # public_access
    if "public_access" in source_properties:
        translated_properties["public_access"] = source_properties["public_access"]

    return translated_properties


def get_translated_resources(
    subscription_id,
    resource_grp_name,
    account_name,
    result_dict,
    original_or_translated,
):
    translated_resource = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_grp_name,
        "account_name": account_name,
        "properties": {},
    }

    if result_dict.get("id"):
        translated_resource["resource_id"] = result_dict.get("id")

    elif result_dict.get("resource_id"):
        translated_resource["resource_id"] = result_dict.get("resource_id")

    if result_dict.get("name"):
        translated_resource["name"] = result_dict.get("name")

    if result_dict.get("type"):
        translated_resource["type"] = result_dict.get("type")

    if result_dict.get("etag"):
        translated_resource["etag"] = result_dict.get("etag")

    if result_dict.get("properties"):
        if original_or_translated:
            translated_resource["properties"] = fill_properties_original(
                translated_resource["properties"], result_dict.get("properties")
            )
        else:
            translated_resource["properties"] = fill_properties_translated(
                translated_resource["properties"], result_dict.get("properties")
            )
    else:
        translated_resource["properties"] = {}
    return translated_resource


def convert_raw_storage_blob_to_present(
    hub,
    subscription_id: str,
    resource_grp_name: str,
    account_name: str,
    result_dict: Dict = None,
) -> Dict[str, Any]:
    return get_translated_resources(
        subscription_id,
        resource_grp_name,
        account_name,
        result_dict,
        original_or_translated=True,
    )


def convert_present_to_raw_storage_blob(
    hub,
    subscription_id: str,
    resource_grp_name: str,
    account_name: str,
    name: str,
    has_immutability_policy: bool = False,
    immutable_storage_with_versioning: bool = False,
    metadata: Dict = None,
    public_access: str = None,
    remaining_retention_days: int = 0,
) -> Dict[str, Any]:
    result_dict = dict()
    if name:
        result_dict["name"] = name
    result_dict["immutable_storage_with_versioning"] = immutable_storage_with_versioning
    translated_resource = get_translated_resources(
        subscription_id,
        resource_grp_name,
        account_name,
        result_dict,
        original_or_translated=True,
    )
    translated_resource["properties"]["deleted"] = False
    translated_resource["properties"][
        "remainingRetentionDays"
    ] = remaining_retention_days
    translated_resource["properties"]["publicAccess"]: public_access
    translated_resource["properties"]["hasImmutabilityPolicy"] = has_immutability_policy
    if metadata:
        translated_resource["properties"]["metadata"] = metadata
    return translated_resource


def convert_raw_storage_blob_to_raw(
    hub,
    subscription_id: str,
    resource_grp_name: str,
    account_name: str,
    result_dict: Dict = None,
) -> Dict[str, Any]:
    return get_translated_resources(
        subscription_id,
        resource_grp_name,
        account_name,
        result_dict,
        original_or_translated=False,
    )


def update_storage_blob_payload(
    hub,
    existing_resources: Dict,
    changed_resources: Dict = None,
) -> Dict[str, Any]:
    result = {"result": True, "ret": None, "comment": []}
    resource_changed = False
    translated_resource = existing_resources

    # default_encryption_scope
    default_encryption_scope_changed = get_changed_resource(
        translated_resource["properties"].get("default_encryption_scope"),
        changed_resources["default_encryption_scope"],
    )
    if default_encryption_scope_changed:
        translated_resource["properties"][
            "default_encryption_scope"
        ] = changed_resources["default_encryption_scope"]
        resource_changed = True

    deny_encryption_scope_override_changed = get_changed_resource(
        translated_resource["properties"].get("deny_encryption_scope_override"),
        changed_resources["deny_encryption_scope_override"],
    )
    if deny_encryption_scope_override_changed:
        translated_resource["properties"][
            "deny_encryption_scope_override"
        ] = changed_resources["deny_encryption_scope_override"]
        resource_changed = True

    has_immutability_policy_changed = get_changed_resource(
        translated_resource["properties"].get("has_immutability_policy"),
        changed_resources["has_immutability_policy"],
    )
    if has_immutability_policy_changed:
        translated_resource["properties"][
            "has_immutability_policy"
        ] = changed_resources["has_immutability_policy"]
        resource_changed = True

    immutable_storage_with_versioning_changed = get_changed_resource(
        translated_resource["properties"].get("has_immutability_policy"),
        changed_resources["has_immutability_policy"],
    )
    if immutable_storage_with_versioning_changed:
        translated_resource["properties"][
            "has_immutability_policy"
        ] = changed_resources["has_immutability_policy"]
        resource_changed = True

    if translated_resource["properties"].get("metadata"):
        metadata_changed = get_changed_resource(
            translated_resource["properties"].get("metadata"),
            changed_resources["metadata"],
        )
        if metadata_changed:
            translated_resource["properties"]["metadata"] = changed_resources[
                "metadata"
            ]
            resource_changed = True
    else:
        translated_resource["properties"]["metadata"] = changed_resources["metadata"]
        resource_changed = True

    public_access_changed = get_changed_resource(
        translated_resource["properties"].get("public_access"),
        changed_resources["public_access"],
    )
    if public_access_changed:
        translated_resource["properties"]["public_access"] = changed_resources[
            "public_access"
        ]
        resource_changed = True

    if resource_changed:
        result["ret"] = translated_resource
    else:
        result["ret"] = None
    return result


def get_changed_resource(original_res, changed_res):
    res = None
    if original_res:
        if changed_res:
            if original_res != changed_res:
                res = changed_res
        else:
            res = changed_res
    else:
        res = changed_res
    return res



"""Exec module for managing SQL databases."""
from collections import OrderedDict
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str, raw: bool = False) -> Dict[str, Any]:
    """Get SQL Database resource by resource_id.

    Args:
        resource_id(str):
            The resource_id of database

        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.sql_database.databases.get resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.sql_database.databases.get
                - kwargs:
                    resource_id: "/subscriptions/11111111-2222-3333-4444-555555555555/resourceGroups/default-rg/providers/Microsoft.Sql/servers/my-server/databases/my-sqldb"
      "

    """
    result = dict(comment=[], result=True, ret=None)

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-11-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].append(response_get["comment"])
        return result

    elif response_get["result"] and response_get["ret"]:
        if raw:
            result["ret"] = response_get["ret"]
        else:
            result[
                "ret"
            ] = hub.tool.azure.sql_database.databases.convert_raw_database_to_present(
                resource=response_get["ret"],
                idem_resource_name=response_get["ret"]["name"],
            )

    return result


async def list_(
    hub, ctx, resource_group_name: str = None, server_name: str = None
) -> Dict:
    """List of SQL databases

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.sql_database.databases.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.sql_database.databases.list
                - kwargs:
                  resource_group_name: default-rg
                  server_name: my-srv
    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id

    if resource_group_name is None or server_name is None:
        async for page_result in hub.tool.azure.request.paginate(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Sql/servers?api-version=2021-11-01",
            success_codes=[200],
        ):
            resource_list = page_result.get("value")
            if resource_list:
                for resource in resource_list:
                    uri_parameters = OrderedDict(
                        {
                            "subscriptions": "subscription_id",
                            "resourceGroups": "resource_group_name",
                            "servers": "server_name",
                        }
                    )
                    uri_parameter_values = (
                        hub.tool.azure.uri.get_parameter_value_in_dict(
                            resource["id"], uri_parameters
                        )
                    )
                    if (
                        resource_group_name is not None
                        and resource_group_name
                        != uri_parameter_values["resource_group_name"]
                    ):
                        continue
                    if (
                        server_name is not None
                        and server_name != uri_parameter_values["server_name"]
                    ):
                        continue
                    result["ret"] += (
                        await _list_by_server(
                            hub,
                            ctx,
                            uri_parameter_values["resource_group_name"],
                            uri_parameter_values["server_name"],
                        )
                    )["ret"]
    else:
        result["ret"] += (
            await _list_by_server(hub, ctx, resource_group_name, server_name)
        )["ret"]

    return result


async def _list_by_server(hub, ctx, resource_group_name: str, server_name: str) -> Dict:
    """List of SQL databases by server

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.sql_database.databases.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.sql_database.databases.list
                - kwargs:
                  resource_group_name: default-rg
                  server_name: my-srv


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id

    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Sql/servers/{server_name}/databases?api-version=2021-11-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                result["ret"].append(
                    hub.tool.azure.sql_database.databases.convert_raw_database_to_present(
                        resource=resource,
                        idem_resource_name=resource["name"],
                    )
                )
        return result



import asyncio
from typing import Any
from typing import Dict
from typing import Optional


async def await_operation(hub, ctx, initial_response, resource_url) -> Dict[str, Any]:
    """
    Awaits a long-running asynchronous Azure operation to complete. Performs an additional GET request when operation completes to fetch the resulting Azure resource.
    :param initial_response: initial HTTP response from Azure API call which initiated the operation
    :param resource_url: Azure API URL, used to GET the resource produced by the operation
    """
    poller = OperationPoller(hub, ctx, initial_response, resource_url)
    return await poller.await_operation()


class OperationFailed(Exception):
    pass


class OperationStatus:
    """Operation status class.

    Operation status is used to indicate the status of an operation. It can be one of the following values: Succeeded, Failed, Canceled, Running.
    """

    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELED = "Canceled"
    RUNNING = "Running"

    FINISHED_STATES = {SUCCEEDED.lower(), FAILED.lower(), CANCELED.lower()}
    FAILED_STATES = {FAILED.lower(), CANCELED.lower()}
    SUCCEEDED_STATES = {SUCCEEDED.lower()}

    @staticmethod
    def finished(status):
        return str(status).lower() in OperationStatus.FINISHED_STATES

    @staticmethod
    def failed(status):
        return str(status).lower() in OperationStatus.FAILED_STATES

    @staticmethod
    def succeeded(status):
        return str(status).lower() in OperationStatus.SUCCEEDED_STATES


class OperationPoller:
    def __init__(self, hub, ctx, initial_response, resource_url):
        self._hub = hub
        self._ctx = ctx
        self._initial_response = initial_response
        self._resource_url = resource_url
        self._status_url = self._get_status_url()
        self._status = None
        self._error = None

    async def await_operation(self) -> Dict[str, Any]:
        return await self._poll()

    def _get_status_url(self) -> Optional[str]:
        headers: Dict[str, Any] = self._initial_response.get("headers")

        if "Azure-AsyncOperation" in headers:
            return headers["Azure-AsyncOperation"]
        elif "Location" in headers:
            return headers["Location"]
        else:
            return None

    def _get_retry_after(self) -> int:
        headers = self._initial_response.get("headers")

        retry_after = headers.get("Retry-After")
        try:
            delay = int(retry_after)
        except ValueError:
            delay = 1

        return delay if delay > 1 else 1

    def _operation_finished(self) -> bool:
        return OperationStatus.finished(self._status)

    def _operation_succeeded(self) -> bool:
        return OperationStatus.succeeded(self._status)

    async def _update_status(self):
        get_response = await self._hub.exec.request.json.get(
            self._ctx,
            url=self._status_url,
            success_codes=[200],
        )
        if "status" in get_response["ret"]:
            self._status = get_response["ret"]["status"].lower()
        else:
            self._status = None
        if "error" in get_response["ret"]:
            self._error = get_response["ret"]["error"]
        else:
            self._error = None

    async def _delay(self):
        await asyncio.sleep(self._get_retry_after())

    async def _poll(self):
        if not self._operation_finished():
            await self._update_status()
        while not self._operation_finished():
            await self._delay()
            await self._update_status()

        if not self._operation_succeeded():
            raise OperationFailed(
                f"Operation failed or has been canceled: {self._error}"
            )
        else:
            get_response = await self._hub.exec.request.json.get(
                self._ctx,
                url=self._resource_url,
                success_codes=[200],
            )
            return get_response



from typing import Any
from typing import Dict
from typing import List


async def paginate(
    hub, ctx, url: str, success_codes: List = [200], headers: Dict = {}
) -> Dict[str, Any]:
    """
    Paginate items from the given azure url
    :param hub: The redistributed pop central hub.
    :param ctx: A dict with the keys/values for the execution of the Idem run located in
     `hub.idem.RUNS[ctx['run_name']]`.
    :param url: HTTP request url
    :param success_codes: List of HTTP status codes that are considered as "successful" to a request
    :param headers: HTTP request headers
    :return: Response body
    """
    while url:
        ret = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=success_codes,
            headers=headers,
        )
        if not ret["result"]:
            raise ValueError(
                f"Error on requesting GET {url} with status code {ret['status']}:"
                f" {ret.get('comment', '')}"
            )
        result = ret["ret"]
        yield result
        url = result.get("nextLink", None)


def _no_update(old_value, new_value) -> bool:
    """Given two value, check if there is a need to update from the old_value to new_value."""
    if isinstance(old_value, str) and isinstance(new_value, str):
        # If the two values are the same string, then no need to update.
        return old_value == new_value
    elif isinstance(old_value, dict) and isinstance(new_value, dict):
        # If the new value dictionary is a subset of the old value dictionary, then no need to update
        return new_value.items() <= old_value.items()
    else:
        return False


def patch_json_content(
    hub,
    patch_parameters: dict,
    old_values: dict,
    new_values: dict,
    force_update_parameters: dict = {"tags": "tags"},
) -> dict:
    """
    Generate a json that contains all the parameters and their values that will be sent during a PATCH operation
    :param hub: The redistributed pop central hub. This is required in
    Idem, so while not used, must appear.
    :param patch_parameters: A dictionary of patchable parameters.
    For example: {"tags": "tags", "properties": {"properties_1": "properties_1", "properties_2": "properties_2"}}
    :param old_values: A dictionary that contains the old values of parameters.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    :param new_values: A dictionary that contains the new values of parameters. This should be the exact structure as
     what old_values have.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    :param force_update_parameters: A dictionary that determines what parameters will be updated as long as the value is
    different, regardless if the new value is a sub-set of old value or not.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    """
    payload = {}
    for parameter_key, parameter_fields in patch_parameters.items():
        value = new_values.get(parameter_key, None)
        if value is None:
            continue
        elif isinstance(parameter_fields, str):
            if parameter_fields in old_values:
                # If the new value is a sub-set of the old value, then we need to decide if this should be a patch
                # operation or a no-op. For example, if tags property changes from {tagA, tagB} to {tagA}, this should
                # trigger a patch operation to update tags property to remove tagB. However, for some properties,
                # seeing a property being absent doesn't mean that we want to delete that property. So to distinguish
                # these two scenarios, we use 'force_update_parameters' to decide what should be updated as long as
                # the value is different.
                if parameter_fields in force_update_parameters:
                    # If a property is in 'force_update_parameters', update the resource as long as old and new value
                    # are different
                    if old_values[parameter_fields] != value:
                        payload.update({parameter_key: value})
                else:
                    # If a property is not in 'force_update_parameters', update the resource only if old and
                    # new value are different and new value is not a sub-set of the old value
                    if old_values[parameter_fields] != value and (
                        not _no_update(old_values[parameter_fields], value)
                    ):
                        payload.update({parameter_key: value})
            else:
                payload.update({parameter_key: value})
        elif isinstance(parameter_fields, dict):
            sub_force_update = force_update_parameters.get(parameter_key, dict())
            if parameter_key in old_values:
                sub_payload = patch_json_content(
                    hub,
                    parameter_fields,
                    old_values[parameter_key],
                    value,
                    sub_force_update,
                )
            else:
                sub_payload = patch_json_content(
                    hub, parameter_fields, dict(), value, sub_force_update
                )
            if sub_payload:
                payload.update({parameter_key: sub_payload})
        else:
            continue
    return payload



import pathlib

import tqdm
from dict_tools.data import NamespaceDict

try:
    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx=ctx, directory=directory)

    ctx.servers = [
        "https://management.azure.com/",
        "https://management.core.windows.net/",
    ]

    # We already have an acct plugin
    ctx.has_acct_plugin = False
    ctx.service_name = ctx.service_name or "azure_auto"
    docs_api = "https://docs.microsoft.com"
    resources = hub.pop_create.azure.resource.parse(
        docs_api, rest_api=ctx.rest_api, resources=ctx.resources
    )

    # Initialize cloud spec
    if ctx.create_plugin == "auto_states":
        request_format = {
            "get": hub.pop_create.azure.template.GET_REQUEST_FORMAT,
            "list": hub.pop_create.azure.template.LIST_REQUEST_FORMAT,
            "create": hub.pop_create.azure.template.CREATE_REQUEST_FORMAT,
            "update": hub.pop_create.azure.template.UPDATE_REQUEST_FORMAT,
            "delete": hub.pop_create.azure.template.DELETE_REQUEST_FORMAT,
        }
    elif ctx.create_plugin == "state_modules":
        request_format = {
            "present": hub.pop_create.azure.template.PRESENT_REQUEST_FORMAT,
            "absent": hub.pop_create.azure.template.ABSENT_REQUEST_FORMAT,
            "describe": hub.pop_create.azure.template.DESCRIBE_REQUEST_FORMAT,
        }

    for ref, resource in tqdm.tqdm(resources.items()):
        try:
            plugins = hub.pop_create.azure.init.plugin(ctx, resource=resource)
            if plugins:
                ctx.cloud_spec = NamespaceDict(
                    project_name=ctx.project_name,
                    service_name=ctx.service_name,
                    request_format=request_format,
                    plugins=plugins,
                )
                hub.cloudspec.init.run(
                    ctx,
                    directory,
                    create_plugins=[ctx.create_plugin],
                )
        except Exception as e:
            hub.log.error(f"Error when generating {ref}: {e.__class__.__name__}: {e}")
        finally:
            ctx.cloud_spec.plugins = {}

    hub.pop_create.init.run(
        directory=directory,
        subparsers=["cicd"],
        **ctx,
    )

    ctx.cloud_spec.plugins = {}
    return ctx


def plugin(hub, ctx, resource):
    plugins = dict()
    for service_name, resources in resource.items():
        service_name_formatted = (
            service_name.lower().strip().replace(" ", "_").replace("-", "_")
        )
        for r in resources:
            resource_name_formatted = (
                r.get("resource").lower().strip().replace(" ", "_").replace("-", "_")
            )
            if not resource_name_formatted:
                continue
            plugin_key = f"{service_name_formatted}.{resource_name_formatted}"

            path = pathlib.Path(ctx.target_directory).absolute() / ctx.clean_name
            skip_existing = not ctx.overwrite_existing
            if skip_existing:
                if ctx.create_plugin == "auto_states":
                    path = path / "exec" / ctx.service_name
                elif ctx.create_plugin == "state_modules":
                    path = path / "states" / ctx.service_name
                elif ctx.create_plugin == "tests":
                    path = path / "tests" / "integration" / "states"

                path = path / service_name_formatted / f"{resource_name_formatted}.py"
                if path.exists():
                    hub.log.info(
                        f"Plugin already exists at '{path}', use `--overwrite` modify"
                    )
                    continue

            shared_function_data = {
                "get_function": f"await hub.exec.request.json.get",
                "create_function": f"await hub.exec.request.json.put",
                "list_function": f"await hub.exec.request.json.get",
                "delete_function": f"await hub.exec.request.raw.delete",
                "patch_function": f"await hub.exec.request.json.patch",
                "resource_name": resource.get("resource"),
                "create_ref": plugin_key,
            }
            try:
                resource_plugin = hub.pop_create.azure.plugin.parse(
                    ctx, r, shared_function_data
                )
            except Exception as e:
                hub.log.error(
                    f"Could not parse {plugin_key}: {e.__class__.__name__}: {e}"
                )
                continue
            plugins[plugin_key] = resource_plugin
    return plugins



from typing import Any
from typing import Dict


RESOURCE_ID_TEMPLATES = {
    "sql_database.databases": "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Sql/servers/{server_name}/databases/{database_name}",
}

RESOURCE_URL_TEMPLATES = {
    "sql_database.databases": "{ctx.acct.endpoint_url}{resource_id}?api-version=2021-11-01",
}


def construct_resource_id(hub, resource_type: str, input_props: Dict[str, Any]) -> str:
    id_template = RESOURCE_ID_TEMPLATES.get(resource_type)
    if not id_template:
        raise ValueError(f"Could not construct resource_id for {resource_type}")

    return id_template.format(**input_props)


def construct_resource_url(
    hub, ctx, resource_type: str, input_props: Dict[str, Any]
) -> str:
    url_template = RESOURCE_URL_TEMPLATES.get(resource_type)
    if not url_template:
        raise ValueError(f"Could not construct resource_url for {resource_type}")

    resource_id = input_props.get(
        "resource_id"
    ) or hub.tool.azure.resource_utils.construct_resource_id(resource_type, input_props)
    subscription_id = input_props.get("subscription_id") or ctx.acct.subscription_id

    return url_template.format(
        **{
            "ctx": ctx,
            **input_props,
            "resource_id": resource_id,
            "subscription_id": subscription_id,
        }
    )


def get_subscription_id_from_account(
    hub, ctx: Dict, subscription_id: str = None
) -> str:
    """If subscription_id is explicitly passed by the user, this subscription_id will be returned.
    If subscription_id is empty, this method will return default subscription_id from Azure account
    :param hub: Hub
    :param ctx: Context for the execution of the Idem run located in `hub.idem.RUNS[ctx['run_name']]`.
    :param subscription_id: A string explicitly passed by the user.
    :return: The correct subscription_id
    """
    if not subscription_id:
        subscription_id = ctx.get("acct", {}).get("subscription_id")
    if not subscription_id:
        hub.log.warning("Could not find subscription_id in account")
    return subscription_id



from typing import Any
from typing import Dict


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    To use mock azure endpoint, provide the endpoint_url in the credentials file.

    Example:
    .. code-block:: yaml

        azure.mock:
          profile_name:
            client_id: Azure client id "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
            secret: Azure client secret "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
            subscription_id: Azure subscription id "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
            tenant: Azure tenant id "cccccccc-cccc-cccc-cccc-cccccccccccc"
            endpoint_url: "azure mock endpoint aaaa.com"
    """
    sub_profiles = {}
    for profile, ctx in profiles.get("azure.mock", {}).items():
        sub_profiles[profile] = ctx
    return sub_profiles
